#!/usr/bin/python
# diskusage.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (diskusage.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: diskusage.

This is OS specific methods to read and check the local disks usage.
Need to be sure user have enough free space on the disk and able to
donate previously specified amount of space.
"""

from __future__ import absolute_import
from __future__ import print_function
import os
import time
import glob

#------------------------------------------------------------------------------

from bitdust.lib import diskspace

from bitdust.main import settings

from bitdust.system import bpio

#------------------------------------------------------------------------------

if bpio.Windows():
    import win32file  # @UnresolvedImport

#------------------------------------------------------------------------------


def GetWinDriveSpace(drive):
    """
    For Windows.

    Return a tuple (<free space in bytes>, <total space in bytes>) or
    (None, None). Call system method ``win32file.GetDiskFreeSpace``.
    """
    try:
        sectorsPerCluster, bytesPerSector, numFreeClusters, totalNumClusters = win32file.GetDiskFreeSpace(drive + ':\\')
        sectorsPerCluster = int(sectorsPerCluster)
        bytesPerSector = int(bytesPerSector)
        numFreeClusters = int(numFreeClusters)
        totalNumClusters = int(totalNumClusters)
    except:
        return None, None
    return float(numFreeClusters*sectorsPerCluster*bytesPerSector), float(totalNumClusters*sectorsPerCluster*bytesPerSector)


def GetLinuxDriveSpace(path):
    """
    For Linux.

    Return a tuple (<free space in bytes>, <total space in bytes>) or
    (None, None). Call system method ``os.statvfs``.
    """
    try:
        s = os.statvfs(str(path))
        # free, total = s.f_bsize*(s.f_blocks-s.f_bavail), s.f_bsize * s.f_bavail
        # free, total = float(s.f_bsize * s.f_bavail), float(s.f_bsize * s.f_blocks)
        # free, total = float(s.f_frsize * s.f_bavail), float(s.f_bsize * s.f_blocks)
        free, total = float(s.f_frsize*s.f_bavail), float(s.f_frsize*s.f_blocks)
        return free, total
    except:
        return None, None


#    if free > total:
#        return total, free
#    else:
#        return free, total


def GetDriveSpace(path):
    """
    So this a sort of portable way to get the free HDD space in the system.
    """
    if bpio.Windows():
        drive = os.path.abspath(path)[0]
        if os.path.isdir(drive + ':'):
            # the drive the data directory is on, ie C
            return GetWinDriveSpace(drive)
        else:
            return None, None
    else:
        # on linux the mount points can make a directory be off a different disk than root
        return GetLinuxDriveSpace(path)


def SumFileSizes(fileList):
    """
    Just iterate the input list and call ``os.path.getsize`` for every item,
    also calculate and return the total size.
    """
    fileSizeTotal = 0
    for filename in fileList:
        try:
            fileSizeTotal += os.path.getsize(filename)
        except:
            pass
    return fileSizeTotal


def GetOurTempFileSizeTotal(
    tempDirectory,
    masks=[
        '*',
    ],
):
    """
    Not used right now.

    Tried here to calculate our temporary files size. Temp files was
    reorganized and so this must be rewritten. TODO.
    """
    ourFileSizes = 0
    for mask in masks:
        ourFileSizes += SumFileSizes(glob.glob(os.path.join(tempDirectory, mask)))
    return ourFileSizes


def OkToShareSpace(desiredSharedSpaceMB):
    """
    Make sure a user really has the space they claim they want to share.
    """
    dataDir = settings.getCustomersFilesDir()
    dataDriveFreeSpace, dataDriveTotalSpace = GetDriveSpace(dataDir)
    if not dataDriveFreeSpace or not dataDriveTotalSpace:
        return False
    try:
        # TODO: say if less than 10% free storage left of your HDD do not share
        testFree = (dataDriveFreeSpace/dataDriveTotalSpace) > 0.1
    except:
        testFree = False
    if not testFree:
        return False
    currentlySharedSpace = GetDirectorySize(dataDir)
    if (currentlySharedSpace + dataDriveFreeSpace/(1024*1024)) < desiredSharedSpaceMB:
        return False
    return True


def GetDirectorySize(directoryPath):
    """
    Calculates the folder size in megabytes using ``bpio.getDirectorySize``.
    """
    return bpio.getDirectorySize(directoryPath)/(1024*1024)


#------------------------------------------------------------------------------


def main():
    """
    This method is for tests.

    Need to move all things here to unit tests. TODO.
    """
    dataDir = settings.getCustomersFilesDir()
    tempDir = settings.TempDir()
    dataDriveFreeSpace = 0
    dataDriveTotalSpace = 0
    tempDriveFreeSpace = 0
    tempDriveTotalSpace = 0

    dataDriveFreeSpace, dataDriveTotalSpace = GetDriveSpace(dataDir)
    tempDriveFreeSpace, tempDriveTotalSpace = GetDriveSpace(tempDir)

    print('data dir =', dataDir)
    print('tep dir =', tempDir)
    print('data dir: ' + str(dataDriveFreeSpace/(1024*1024)) + 'MB free/' + str(dataDriveTotalSpace/(1024*1024)) + 'MB total')
    print('temp dir: ' + str(tempDriveFreeSpace/(1024*1024)) + 'MB free/' + str(tempDriveTotalSpace/(1024*1024)) + 'MB total')

    print(time.time())
    print('our temp files: ' + str(GetOurTempFileSizeTotal(tempDir)/(1024*1024)) + 'MB')
    print(time.time())

    GetDirectorySize(dataDir)

    ds = diskspace.DiskSpace()
    print(ds.getValueBest(dataDriveFreeSpace))

    print('at OkToShareSpace ...')
    print('ok to share 100MB - should be true')
    print(OkToShareSpace(100))
    print('ok to share 12345678MB - should be false')
    print(OkToShareSpace(12345678))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# packet_in.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (packet_in.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: packet_in.

.. role:: red

BitDust packet_in() Automat

.. raw:: html

    <a href="packet_in.png" target="_blank">
    <img src="packet_in.png" style="max-width:100%;">
    </a>

EVENTS:
    * :red:`cancel`
    * :red:`failed`
    * :red:`register-item`
    * :red:`remote-id-cached`
    * :red:`unregister-item`
    * :red:`unserialize-failed`
    * :red:`valid-inbox-packet`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 16

_PacketLogFileEnabled = False

#------------------------------------------------------------------------------

import os
import time

from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.main import settings
from bitdust.main import config

from bitdust.automats import automat

from bitdust.lib import nameurl
from bitdust.lib import strng

from bitdust.system import bpio
from bitdust.system import tmpfile

from bitdust.userid import global_id
from bitdust.userid import id_url

from bitdust.contacts import contactsdb
from bitdust.contacts import identitycache

from bitdust.services import driver

from bitdust.p2p import commands
from bitdust.p2p import p2p_stats

from bitdust.transport import callback

#------------------------------------------------------------------------------

_InboxItems = {}
_PacketsCounter = 0
_History = []

#------------------------------------------------------------------------------


def init():
    global _PacketLogFileEnabled
    _PacketLogFileEnabled = config.conf().getBool('logs/packet-enabled')


def shutdown():
    global _PacketLogFileEnabled
    _PacketLogFileEnabled = False


#------------------------------------------------------------------------------


def get_packets_counter():
    global _PacketsCounter
    return _PacketsCounter


def increment_packets_counter():
    global _PacketsCounter
    _PacketsCounter += 1


#------------------------------------------------------------------------------


def inbox_items():
    global _InboxItems
    return _InboxItems


def create(transfer_id):
    p = PacketIn(transfer_id)
    inbox_items()[transfer_id] = p
    # lg.out(10, 'packet_in.create  %s,  %d working items now' % (
    #     transfer_id, len(items())))
    return p


def get(transfer_id):
    return inbox_items().get(transfer_id, None)


def search(sender_idurl=None, proto=None, host=None):
    """
    Returns list of transfer ids of incoming packets which satisfies given criteria.
    """
    if sender_idurl and not isinstance(sender_idurl, list):
        sender_idurl = [
            sender_idurl,
        ]
    results = set()
    for transfer_id, itm in inbox_items().items():
        if sender_idurl:
            if itm.sender_idurl and itm.sender_idurl == sender_idurl:
                results.add(transfer_id)
                continue
        if proto and host:
            if itm.proto and itm.proto == proto and itm.host and itm.host == host:
                results.add(transfer_id)
                continue
        if host:
            if itm.host and itm.host == host:
                results.add(transfer_id)
                continue
        if proto:
            if itm.proto and itm.proto == proto:
                results.add(transfer_id)
                continue
    return list(results)


def history():
    global _History
    return _History


#------------------------------------------------------------------------------


def process(newpacket, info):
    """
    Main entry point where all incoming signed packets are coming from remote peers.
    The main aspect here is to "authenticate" remote node - need to know it identity.
    """
    # from bitdust.p2p import p2p_service
    from bitdust.transport import gateway
    from bitdust.userid import my_id
    if not driver.is_on('service_p2p_hookups'):
        if _Debug:
            lg.out(_DebugLevel, 'packet_in.process SKIP incoming packet, service_p2p_hookups is not started')
        return None
    if _Debug:
        lg.out(
            _DebugLevel, 'packet_in.process [%s/%s/%s]:%s(%s) from %s://%s is "%s"' % (
                nameurl.GetName(newpacket.OwnerID),
                nameurl.GetName(newpacket.CreatorID),
                nameurl.GetName(newpacket.RemoteID),
                newpacket.Command,
                newpacket.PacketID,
                info.proto,
                info.host,
                info.status,
            )
        )
    if info.status != 'finished':
        if _Debug:
            lg.out(_DebugLevel, '    skip, packet status is : [%s]' % info.status)
        return None


#     if _PacketLogFileEnabled:
#         lg.out(0, '        \033[0;49;92mIN %s(%s) with %d bytes from %s to %s TID:%s\033[0m' % (
#             newpacket.Command, newpacket.PacketID, info.bytes_received,
#             global_id.UrlToGlobalID(info.sender_idurl), global_id.UrlToGlobalID(newpacket.RemoteID),
#             info.transfer_id), log_name='packet', showtime=True)
# we must know recipient identity
    if not id_url.is_cached(newpacket.RemoteID):
        d = identitycache.immediatelyCaching(newpacket.RemoteID)
        d.addCallback(lambda _: process(newpacket, info))
        d.addErrback(lambda err: lg.err('incoming remote ID is unknown, failed caching remote %s identity: %s' % (newpacket.RemoteID, str(err))) and None)
        return d
    if newpacket.Command == commands.Identity():
        if not id_url.is_the_same(newpacket.RemoteID, my_id.getIDURL()):
            if _Debug:
                lg.out(_DebugLevel, '    incoming Identity is routed to another user')
            if not gateway.on_identity_received(newpacket, send_ack=False):
                lg.warn('received identity was not processed')
                return None
            # remote peer sending a valid identity to another peer routed via my machine
            # need to handle that packet - it should be processed by proxy_server
            return handle(newpacket, info)
        # contact sending us current identity we might not have
        # so we handle it before check that packet is valid
        # because we might not have his identity on hands and so can not verify the packet
        # so we check that his Identity is valid and save it into cache
        # than we check the packet to be valid too.
        if not gateway.on_identity_received(newpacket):
            lg.warn('received identity was not processed')
            return None
    if not identitycache.HasKey(newpacket.CreatorID):
        if _Debug:
            lg.out(_DebugLevel, '    will cache remote identity %s before processing incoming packet %s' % (newpacket.CreatorID, newpacket))
        d = identitycache.immediatelyCaching(newpacket.CreatorID)
        d.addCallback(lambda _: handle(newpacket, info))
        d.addErrback(lambda err: lg.err('failed caching remote %s identity: %s' % (newpacket.CreatorID, str(err))) and None)
        return d
    return handle(newpacket, info)


def handle(newpacket, info):
    """
    Actually process incoming packet. Here we can be sure that owner/creator of the packet is identified.
    """
    from bitdust.transport import packet_out
    handled = False
    # check that signed by a contact of ours
    try:
        is_signature_valid = newpacket.Valid(raise_signature_invalid=False)
    except:
        is_signature_valid = False
        # lg.exc('new packet from %s://%s is NOT VALID:\n\n%r\n' % (
        #     info.proto, info.host, newpacket.Serialize()))
    if not is_signature_valid:
        if _Debug:
            lg.args(_DebugLevel, PacketID=newpacket.PacketID, OwnerID=newpacket.OwnerID, CreatorID=newpacket.CreatorID, RemoteID=newpacket.RemoteID)
        lg.warn('signature is not valid for %r from %r|%r to %r' % (newpacket, newpacket.OwnerID, newpacket.CreatorID, newpacket.RemoteID))
        return None
    try:
        if not commands.IsRelay(newpacket.Command):
            for p in packet_out.search_by_response_packet(newpacket, info.proto, info.host):
                p.automat('inbox-packet', (newpacket, info))
                handled = True
                if _Debug:
                    lg.out(_DebugLevel, '    processed by %s as response packet' % p)
    except:
        lg.exc()
    try:
        handled = callback.run_inbox_callbacks(newpacket, info, info.status, info.error_message) or handled
    except:
        lg.exc()
    if not handled and newpacket.Command not in [
        commands.Ack(),
        commands.Fail(),
        commands.Identity(),
    ]:
        lg.warn('incoming %s from [%s://%s] was NOT HANDLED' % (newpacket, info.proto, info.host))
        if _PacketLogFileEnabled:
            lg.out(
                0,
                '                \033[1;49;91mIN NOT HANDLED %s(%s) with %d bytes from %s to %s TID:%s\033[0m' % (
                    newpacket.Command,
                    newpacket.PacketID,
                    info.bytes_received,
                    global_id.UrlToGlobalID(info.sender_idurl),
                    global_id.UrlToGlobalID(newpacket.RemoteID),
                    info.transfer_id,
                ),
                log_name='packet',
                showtime=True,
            )
    else:
        if _PacketLogFileEnabled:
            lg.out(
                0,
                '                \033[0;49;93mIN OK %s(%s) with %d bytes from %s to %s TID:%s\033[0m' % (
                    newpacket.Command,
                    newpacket.PacketID,
                    info.bytes_received,
                    global_id.UrlToGlobalID(info.sender_idurl),
                    global_id.UrlToGlobalID(newpacket.RemoteID),
                    info.transfer_id,
                ),
                log_name='packet',
                showtime=True,
            )
    if _Debug and False:
        history().append(
            {
                'time': newpacket.Date,
                'command': newpacket.Command,
                'packet_id': newpacket.PacketID,
                'creator_id': newpacket.CreatorID,
                'owner_id': newpacket.OwnerID,
                'remote_id': newpacket.RemoteID,
                'payload': len(newpacket.Payload),
                'address': '%s://%s' % (info.proto, info.host),
            }
        )
        if len(history()) > 100:
            history().pop(0)
    return handled


#------------------------------------------------------------------------------


class PacketIn(automat.Automat):
    """
    This class implements all the functionality of the ``packet_in()`` state
    machine.
    """
    def __init__(self, transfer_id):
        self.transfer_id = transfer_id
        self.time = None
        self.timeout = None
        self.proto = None
        self.host = None
        self.sender_idurl = None
        self.filename = None
        self.size = None
        self.bytes_received = None
        self.status = None
        self.error_message = None
        _counter = get_packets_counter()
        increment_packets_counter()
        self.label = ''
        automat.Automat.__init__(
            self,
            name='in_%d_%s' % (_counter, self.transfer_id),
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
            publish_events=False,
        )

    def __repr__(self):
        """
        Will return something like: "in_4_1315435345[Ack(987654321)](DONE)".
        """
        return '%s%s(%s)' % (self.id, self.label, self.state)

    def is_timed_out(self):
        #         if self.time is None or self.timeout is None:
        #             return False
        #         return time.time() - self.time > self.timeout
        return False

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'register-item':
                self.state = 'RECEIVING'
                self.doInit(*args, **kwargs)
        #---RECEIVING---
        elif self.state == 'RECEIVING':
            if event == 'cancel':
                self.doCancelItem(*args, **kwargs)
            elif event == 'unregister-item' and not self.isTransferFinished(*args, **kwargs):
                self.state = 'FAILED'
                self.doReportFailed(*args, **kwargs)
                self.doEraseInputFile(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'unregister-item' and self.isTransferFinished(*args, **kwargs) and not self.isRemoteIdentityCached(*args, **kwargs):
                self.state = 'CACHING'
                self.doCacheRemoteIdentity(*args, **kwargs)
            elif event == 'unregister-item' and self.isTransferFinished(*args, **kwargs) and self.isRemoteIdentityCached(*args, **kwargs):
                self.state = 'INBOX?'
                self.doReadAndUnserialize(*args, **kwargs)
        #---INBOX?---
        elif self.state == 'INBOX?':
            if event == 'valid-inbox-packet':
                self.state = 'DONE'
                self.doReportReceived(*args, **kwargs)
                self.doEraseInputFile(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'unserialize-failed':
                self.state = 'FAILED'
                self.doReportFailed(*args, **kwargs)
                self.doEraseInputFile(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---FAILED---
        elif self.state == 'FAILED':
            pass
        #---DONE---
        elif self.state == 'DONE':
            pass
        #---CACHING---
        elif self.state == 'CACHING':
            if event == 'failed':
                self.state = 'FAILED'
                self.doReportCacheFailed(*args, **kwargs)
                self.doEraseInputFile(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'remote-id-cached':
                self.state = 'INBOX?'
                self.doReadAndUnserialize(*args, **kwargs)
        return None

    def isTransferFinished(self, *args, **kwargs):
        """
        Condition method.
        """
        status, bytes_received, _ = args[0]
        if status != 'finished':
            return False
        if self.size and self.size > 0 and self.size != bytes_received:
            return False
        return True

    def isRemoteIdentityCached(self, *args, **kwargs):
        """
        Condition method.
        """
        if not self.sender_idurl:
            return True
        return self.sender_idurl and identitycache.HasKey(self.sender_idurl)

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.proto, self.host, self.sender_idurl, self.filename, self.size = args[0]
        self.time = time.time()
        # 300  # max(10 * int(self.size/float(settings.SendingSpeedLimit())), 10)
        if self.size < 1024*10:
            self.timeout = 10
        elif self.size > 1024*1024:
            self.timeout = int(self.size/float(settings.SendingSpeedLimit()))
        else:
            self.timeout = 300
        if not self.sender_idurl:
            lg.warn('sender_idurl is None: %s' % str(*args, **kwargs))
        reactor.callLater(0, callback.run_begin_file_receiving_callbacks, self)  # @UndefinedVariable

    def doEraseInputFile(self, *args, **kwargs):
        """
        Action method.
        """
        reactor.callLater(1, tmpfile.throw_out, self.filename, 'received')  # @UndefinedVariable

    def doCancelItem(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport import gateway
        t = gateway.transports().get(self.proto, None)
        if t:
            t.call('cancel_file_receiving', self.transfer_id)

    def doCacheRemoteIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        d = identitycache.immediatelyCaching(self.sender_idurl)
        d.addCallback(self._on_remote_identity_cached, *args, **kwargs)
        d.addErrback(self._on_remote_identity_cache_failed, *args, **kwargs)

    def doReadAndUnserialize(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.transport import gateway
        self.status, self.bytes_received, self.error_message = args[0]
        if _PacketLogFileEnabled:
            lg.out(
                0,
                '                \033[2;49;32mRECEIVED %d bytes from %s://%s TID:%s\033[0m' % (self.bytes_received, self.proto, self.host, self.transfer_id),
                log_name='packet',
                showtime=True,
            )
        # DO UNSERIALIZE HERE , no exceptions
        newpacket = gateway.inbox(self)
        if newpacket is None:
            if _Debug:
                lg.out(_DebugLevel, '<<< IN <<< !!!NONE!!! [%s] %s from %s %s' % (
                    self.proto.upper().ljust(5),
                    self.status.ljust(8),
                    self.host,
                    os.path.basename(self.filename),
                ))
            # net_misc.ConnectionFailed(None, proto, 'receiveStatusReport %s' % host)
            try:
                fd, _ = tmpfile.make('error', extension='.inbox')
                data = bpio.ReadBinaryFile(self.filename)
                os.write(fd, strng.to_bin('from %s:%s %s\n' % (self.proto, self.host, self.status)))
                os.write(fd, data)
                os.close(fd)
            except:
                lg.exc()
            if os.path.isfile(self.filename):
                try:
                    os.remove(self.filename)
                except:
                    lg.exc()
            self.automat('unserialize-failed', None)
            return
        self.label = '[%s@%s]' % (newpacket.Command, newpacket.PacketID)
        if _Debug:
            lg.out(_DebugLevel, 'packet_in.doReadAndUnserialize: %s' % newpacket)
        self.automat('valid-inbox-packet', newpacket)

    def doReportReceived(self, *args, **kwargs):
        """
        Action method.
        """
        newpacket = args[0]
        p2p_stats.count_inbox(self.sender_idurl, self.proto, self.status, self.bytes_received)
        process(newpacket, self)

    def doReportFailed(self, *args, **kwargs):
        """
        Action method.
        """
        try:
            status, bytes_received, _ = args[0]
        except:
            status = 'failed'
            bytes_received = 0
        p2p_stats.count_inbox(self.sender_idurl, self.proto, status, bytes_received)
        lg.warn('incoming packet failed %s with %s' % (self.transfer_id, status))
        if _PacketLogFileEnabled:
            lg.out(
                0,
                '                \033[0;49;31mIN FAILED with status "%s" from %s://%s TID:%s\033[0m' % (status, self.proto, self.host, self.transfer_id),
                log_name='packet',
                showtime=True,
            )

    def doReportCacheFailed(self, *args, **kwargs):
        """
        Action method.
        """
        if args and args[0]:
            status, bytes_received, msg = args[0]
            p2p_stats.count_inbox(self.sender_idurl, self.proto, status, bytes_received)
        else:
            status = 'failed'
            bytes_received = 0
            msg = 'unknown reason'
        lg.warn('cache failed : %s' % self.sender_idurl)
        if _PacketLogFileEnabled:
            lg.out(
                0,
                '                \033[0;49;31mIN CACHE FAILED with "%s" for %s TID:%s\033[0m' % (msg, self.sender_idurl, self.transfer_id),
                log_name='packet',
                showtime=True,
            )

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        inbox_items().pop(self.transfer_id)
        self.destroy()

    def _on_remote_identity_cached(self, xmlsrc, *args, **kwargs):
        sender_identity = contactsdb.get_contact_identity(self.sender_idurl)
        if sender_identity is None:
            self.automat('failed')
        else:
            self.automat('remote-id-cached', *args, **kwargs)

    def _on_remote_identity_cache_failed(self, err, *args, **kwargs):
        lg.warn('%s : %s' % (repr(self), str(err)))
        self.automat('failed')
        return None



import select, json, platform
import threading

# Logical timeout
LTIMEOUT = 15
# Fixed header length
SLEN = 10

_Debug = False


def send(sdef, data, slen=SLEN):
    sdef.setblocking(1)
    # Make sure the packet is sent in one call
    j = json.dumps(data)
    if _Debug:
        print(f'  connections.send {sdef.getsockname()} {sdef.getpeername()} <{j[:12]}> in {threading.current_thread().name}')
    return sdef.sendall(str(len(j)).encode('utf-8').zfill(slen) + j.encode('utf-8'))


if 'Linux' in platform.system():
    READ_OR_ERROR = select.POLLIN | select.POLLPRI | select.POLLHUP | select.POLLERR | select.POLLNVAL

    #READ_ONLY = select.POLLIN | select.POLLPRI


    def receive(sdef, slen=SLEN, timeout=LTIMEOUT):
        try:
            sdef.setblocking(1)
            poller = select.poll()
            poller.register(sdef, READ_OR_ERROR)
            ready = poller.poll(timeout*1000)
            if not ready:
                # logical timeout
                return '*'
            fd, flag = ready[0]
            if (flag & (select.POLLHUP | select.POLLERR | select.POLLNVAL)):
                # No need to read
                raise RuntimeError('Socket POLLHUP')
            if (flag & (select.POLLIN | select.POLLPRI)):
                data = sdef.recv(slen)
                if not data:
                    # POLLIN and POLLHUP are not exclusive. We can have both.
                    raise RuntimeError('Socket EOF')
                data = int(data)  # receive length
            elif (flag & (select.POLLERR | select.POLLHUP | select.POLLNVAL)):
                raise RuntimeError('Socket error {}'.format(flag))
            else:
                raise RuntimeError('Socket Unexpected Error')
            chunks = []
            bytes_recd = 0
            while bytes_recd < data:
                ready = poller.poll(timeout*1000)
                if not ready:
                    raise RuntimeError('Socket Timeout2')
                fd, flag = ready[0]
                if (flag & (select.POLLHUP | select.POLLERR | select.POLLNVAL)):
                    # No need to read
                    raise RuntimeError('Socket POLLHUP2')
                if (flag & (select.POLLIN | select.POLLPRI)):
                    chunk = sdef.recv(min(data - bytes_recd, 2048))
                    if not chunk:
                        raise RuntimeError('Socket EOF2')
                    chunks.append(chunk)
                    bytes_recd = bytes_recd + len(chunk)
                elif (flag & (select.POLLERR | select.POLLHUP | select.POLLNVAL)):
                    raise RuntimeError('Socket Error {}'.format(flag))
                else:
                    raise RuntimeError('Socket Unexpected Error')

            poller.unregister(sdef)
            segments = b''.join(chunks).decode('utf-8')
            ret = json.loads(segments)
            if _Debug:
                print(f'  connections.receive {sdef.getsockname()} {sdef.getpeername()} <{segments[:10]}> in {threading.current_thread().name}')
            return ret

        except Exception as e:
            if _Debug:
                print('        connections.receive Exception:', sdef.getsockname(), sdef.getpeername(), e)
            """
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print(exc_type, fname, exc_tb.tb_lineno)
            """
            # Final cleanup
            try:
                poller.unregister(sdef)
            except Exception as e2:
                pass
                #print ("Exception unregistering: {}".format(e2))
            raise RuntimeError(f'Connections: {e}')

else:

    def receive(sdef, slen=SLEN, timeout=LTIMEOUT):
        sdef.setblocking(1)
        ready = select.select([sdef], [], [sdef], timeout)
        if ready[0]:
            try:
                data = int(sdef.recv(slen))  # receive length
                #print ("To receive: {}".format(data))
            except:
                raise RuntimeError('Connection closed by the remote host')

        else:
            # logical timeoutsha_hash
            return '*'
            #raise RuntimeError("Socket timeout")

        chunks = []
        bytes_recd = 0
        while bytes_recd < data:
            ready = select.select([sdef], [], [], timeout)
            if ready[0]:
                chunk = sdef.recv(min(data - bytes_recd, 2048))
                if not chunk:
                    raise RuntimeError('Socket connection broken')
                chunks.append(chunk)
                bytes_recd = bytes_recd + len(chunk)
            else:
                raise RuntimeError('Socket timeout')

        segments = b''.join(chunks).decode('utf-8')
        #print(f"Received segments: {segments} from {sdef.getpeername()[0]}")

        return json.loads(segments)



#!/usr/bin/env python
# id_rotator.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (id_rotator.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
"""
.. module:: id_rotator
.. role:: red

.. raw:: html

    <a href="https://bitdust.io/wiki/p2p/id_rotator.png" target="_blank">
    <img src="https://bitdust.io/wiki/p2p/id_rotator.png" style="max-width:100%;">
    </a>


BitDust id_rotator() Automat

EVENTS:
    * :red:`auto-rotate-disabled`
    * :red:`check`
    * :red:`found-new-id-source`
    * :red:`id-server-failed`
    * :red:`my-id-exist`
    * :red:`my-id-failed`
    * :red:`my-id-sent`
    * :red:`my-id-updated`
    * :red:`need-more-sources`
    * :red:`no-id-servers-found`
    * :red:`ping-done`
    * :red:`ping-failed`
    * :red:`run`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import sys as _s, os.path as _p
    _s.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(_s.argv[0])), '..')))

#------------------------------------------------------------------------------

import random

from twisted.internet.defer import Deferred, DeferredList, CancelledError  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import net_misc
from bitdust.lib import nameurl
from bitdust.lib import strng

from bitdust.main import config
from bitdust.main import settings

from bitdust.userid import identity
from bitdust.userid import known_servers
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_IdRotator = None

#------------------------------------------------------------------------------


def check():
    """
    Returns True/False via deferred object to report situation with my identity sources.
    """
    result_defer = Deferred()
    ir = IdRotator()
    ir.automat('check', result_defer=result_defer)
    return result_defer


def run(preferred_servers={}, force=False):
    """
    Executes id_rotator() state machine to test my identity sources and if needed republish my identity
    on another ID server in case current situation with ID sources "is not healthy".
    Input parameter "preferred_servers" can be used to control which servers to be used
    when replacing a dead one.
    Returns True/False via deferred object to report result.
    """
    result_defer = Deferred()
    ir = IdRotator()
    ir.automat('run', result_defer=result_defer, preferred_servers=preferred_servers, force=force)
    return result_defer


#------------------------------------------------------------------------------


class IdRotator(automat.Automat):
    """
    This class implements all the functionality of ``id_rotator()`` state machine.
    """
    def __init__(self, debug_level=_DebugLevel, log_events=_Debug, log_transitions=_Debug, publish_events=True, **kwargs):
        """
        Builds `id_rotator()` state machine.
        """
        super(IdRotator, self).__init__(name='id_rotator', state='AT_STARTUP', debug_level=debug_level, log_events=log_events, log_transitions=log_transitions, publish_events=publish_events, **kwargs)

    def init(self):
        """
        Method to initialize additional variables and flags
        at creation phase of `id_rotator()` machine.
        """

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <http://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'run' or event == 'check':
                self.state = 'ID_SERVERS?'
                self.doInit(event, *args, **kwargs)
                self.doPingMyIDServers(*args, **kwargs)
        #---ID_SERVERS?---
        elif self.state == 'ID_SERVERS?':
            if event == 'ping-failed':
                self.state = 'FAILED'
                self.doReportFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'ping-done' and (self.isChecking(*args, **kwargs) or self.isHealthy(*args, **kwargs)):
                self.state = 'DONE'
                self.doReportDone(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'ping-done' and not self.isChecking(*args, **kwargs) and not self.isHealthy(*args, **kwargs):
                self.state = 'NEW_SOURCE!'
                self.doSelectNewIDServer(*args, **kwargs)
        #---NEW_SOURCE!---
        elif self.state == 'NEW_SOURCE!':
            if event == 'found-new-id-source':
                self.state = 'MY_ID_ROTATE'
                self.doRebuildMyIdentity(*args, **kwargs)
            elif event == 'no-id-servers-found' or event == 'auto-rotate-disabled':
                self.state = 'FAILED'
                self.doReportFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---MY_ID_ROTATE---
        elif self.state == 'MY_ID_ROTATE':
            if event == 'my-id-updated':
                self.state = 'SEND_ID'
                self.doSendMyIdentity(*args, **kwargs)
            elif event == 'need-more-sources':
                self.state = 'NEW_SOURCE!'
                self.doSelectNewIDServer(*args, **kwargs)
        #---SEND_ID---
        elif self.state == 'SEND_ID':
            if event == 'my-id-sent':
                self.state = 'REQUEST_ID'
                self.doRequestMyIdentity(*args, **kwargs)
            elif event == 'my-id-failed':
                self.state = 'FAILED'
                self.doReportFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---REQUEST_ID---
        elif self.state == 'REQUEST_ID':
            if event == 'my-id-exist' and self.isMyIdentityValid(*args, **kwargs):
                self.state = 'DONE'
                self.doReportDone(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'id-server-failed' or (event == 'my-id-exist' and not self.isMyIdentityValid(*args, **kwargs)):
                self.state = 'FAILED'
                self.doReportFailed(event, *args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---DONE---
        elif self.state == 'DONE':
            pass
        #---FAILED---
        elif self.state == 'FAILED':
            pass
        return None

    def isHealthy(self, *args, **kwargs):
        """
        Condition method.
        """
        if self.force:
            # this way we can execute "rotate" flow even if all identity sources are healthy
            return False
        return self._is_healthy(args[0])

    def isChecking(self, *args, **kwargs):
        """
        Condition method.
        """
        return self.check_only

    def isMyIdentityValid(self, *args, **kwargs):
        """
        Condition method.
        """
        # TODO: check all results here
        return True

    def doInit(self, event, *args, **kwargs):
        """
        Action method.
        """
        self.result_defer = kwargs.get('result_defer')
        self.check_only = False
        if event == 'check':
            self.check_only = True
        self.alive_idurls = []
        self.old_sources = my_id.getLocalIdentity().getSources(as_originals=True)
        self.known_servers = known_servers.by_host()
        self.preferred_servers = kwargs.get('preferred_servers', {})
        self.possible_sources = []
        if _Debug:
            lg.args(_DebugLevel, preferred_servers=self.preferred_servers)
        self.force = kwargs.get('force', False)
        self.new_revision = kwargs.get('new_revision')
        self.rotated = False
        if not self.preferred_servers:
            try:
                for srv in strng.to_text(config.conf().getString('services/identity-propagate/known-servers')).split(','):
                    if srv.strip():
                        parts = srv.strip().split(':')
                        if len(parts) == 2:
                            host, web_port = parts
                            tcp_port = settings.IdentityServerPort()
                        else:
                            host, web_port, tcp_port = parts
                        self.preferred_servers[host] = (
                            int(web_port),
                            int(tcp_port),
                        )
            except:
                lg.exc()
        self.preferred_servers = {strng.to_bin(k): v for k, v in self.preferred_servers.items()}
        self.current_servers = []
        for idurl in my_id.getLocalIdentity().getSources(as_originals=True):
            self.current_servers.append(strng.to_bin(nameurl.GetHost(idurl)))
        if _Debug:
            lg.args(_DebugLevel, known_servers=self.known_servers, preferred_servers=self.preferred_servers, current_servers=self.current_servers)

    def doPingMyIDServers(self, *args, **kwargs):
        """
        Action method.
        """
        self._do_verify_my_sources()

    def doSelectNewIDServer(self, *args, **kwargs):
        """
        Action method.
        """
        if not self.force and not config.conf().getBool('services/identity-propagate/automatic-rotate-enabled'):
            self.automat('auto-rotate-disabled')
            return

        target_servers = self.preferred_servers
        if not target_servers:
            target_servers = self.known_servers
        # make sure to not choose a server I already have in my ID sources
        for current_server in self.current_servers:
            if current_server in target_servers:
                target_servers.pop(current_server)

        if not target_servers:
            self.automat('no-id-servers-found')
            return

        target_servers = {strng.to_bin(k): v for k, v in target_servers.items()}
        target_hosts = list(target_servers.keys())
        random.shuffle(target_hosts)
        if _Debug:
            lg.args(_DebugLevel, target_hosts=target_hosts, current_servers=self.current_servers, target_servers=target_servers)

        def _new_idurl_exist(idsrc, new_idurl, pos):
            if _Debug:
                lg.out(_DebugLevel, 'id_rotator.doSelectNewIDServer._new_idurl_exist %r already with same name' % new_idurl)
            latest_revision = my_id.getLocalIdentity().getRevisionValue()
            try:
                existing_identity_with_same_name = identity.identity(xmlsrc=idsrc)
                if not existing_identity_with_same_name.isCorrect():
                    raise Exception('remote identity not correct at position %r' % pos)
                if not existing_identity_with_same_name.Valid():
                    raise Exception('remote identity not valid at position %r' % pos)
            except:
                lg.exc()
                if pos + 1 >= len(target_hosts):
                    self.automat('no-id-servers-found')
                else:
                    _ping_one_server(pos + 1)
                return
            if existing_identity_with_same_name.getPublicKey() == my_id.getLocalIdentity().getPublicKey():
                if latest_revision <= existing_identity_with_same_name.getRevisionValue():
                    self.new_revision = max(self.new_revision or -1, existing_identity_with_same_name.getRevisionValue() + 1)
                lg.info('found my own identity on "old" ID server and will re-use that source again: %r' % new_idurl)
                if new_idurl not in self.possible_sources:
                    self.possible_sources.append(new_idurl)
                self.automat('found-new-id-source', new_idurl)
                return

            if pos + 1 >= len(target_hosts):
                self.automat('no-id-servers-found')
            else:
                _ping_one_server(pos + 1)

        def _new_idurl_not_exist(err, new_idurl):
            lg.info('found new identity source available to use: %r' % new_idurl)
            if new_idurl not in self.possible_sources:
                self.possible_sources.append(new_idurl)
            self.automat('found-new-id-source', new_idurl)

        def _server_replied(htmlsrc, host, pos):
            if _Debug:
                lg.out(_DebugLevel, 'id_rotator.doSelectNewIDServer._server_replied %r' % host)
            webport, _ = target_servers[host]
            if webport == 80:
                webport = ''
            new_idurl = nameurl.UrlMake('http', strng.to_text(host), webport, my_id.getIDName() + '.xml')
            d = net_misc.getPageTwisted(new_idurl, timeout=15)
            d.addCallback(_new_idurl_exist, new_idurl, pos)
            d.addErrback(_new_idurl_not_exist, new_idurl)

        def _server_failed(err, host, pos):
            if _Debug:
                lg.out(_DebugLevel, 'id_rotator.doSelectNewIDServer._server_failed %r with %r' % (host, err))
            if pos + 1 >= len(target_hosts):
                self.automat('no-id-servers-found')
            else:
                _ping_one_server(pos + 1)

        def _ping_one_server(pos):
            host = target_hosts[pos]
            webport, tcpport = target_servers[host]
            if webport == 80:
                webport = ''
            tcpport = int(tcpport)
            server_url = nameurl.UrlMake('http', strng.to_text(host), webport, '')
            if _Debug:
                lg.out(_DebugLevel, 'id_rotator.doSelectNewIDServer._ping_one_server at %s known tcp port is %d' % (server_url, tcpport))
            d = net_misc.getPageTwisted(server_url, timeout=15)
            d.addCallback(_server_replied, host, pos)
            d.addErrback(_server_failed, host, pos)

        _ping_one_server(0)

    def doRebuildMyIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        min_servers = max(
            settings.MinimumIdentitySources(),
            config.conf().getInt('services/identity-propagate/min-servers') or settings.MinimumIdentitySources(),
        )
        max_servers = min(
            settings.MaximumIdentitySources(),
            config.conf().getInt('services/identity-propagate/max-servers') or settings.MaximumIdentitySources(),
        )
        current_sources = my_id.getLocalIdentity().getSources(as_originals=True)
        current_contacts = list(my_id.getLocalIdentity().getContacts())
        new_sources = []
        new_idurl = strng.to_bin(args[0])
        if _Debug:
            lg.args(_DebugLevel, current_sources=current_sources, alive_idurls=self.alive_idurls, new_idurl=new_idurl)
        # first get rid of "dead" sources
        for current_idurl in current_sources:
            if current_idurl not in self.alive_idurls:
                continue
            if strng.to_bin(current_idurl) in new_sources:
                continue
            new_sources.append(strng.to_bin(current_idurl))
        if self.force and len(new_sources) == len(current_sources):
            # do not increase number of identity sources, only rotate them
            new_sources.pop(0)
        # and add new "good" source to the end of the list
        if new_idurl and new_idurl not in new_sources:
            new_sources.append(new_idurl)
        if _Debug:
            lg.args(_DebugLevel, new_sources=new_sources, min_servers=min_servers, max_servers=max_servers)
        if len(new_sources) > max_servers:
            all_new_sources = list(new_sources)
            new_sources = new_sources[max(0, len(new_sources) - max_servers):]
            lg.warn('skip %d identity sources, require maximum %d sources' % (len(all_new_sources) - len(new_sources), max_servers))
        if len(new_sources) < min_servers:
            additional_sources = self.possible_sources[:min_servers - len(new_sources)]
            if additional_sources:
                lg.warn('additional sources to be used: %r' % additional_sources)
                new_sources.extend(additional_sources)
        unique_sources = []
        for idurl_bin in new_sources:
            if strng.to_bin(idurl_bin) not in unique_sources:
                unique_sources.append(strng.to_bin(idurl_bin))
        if len(unique_sources) < min_servers:
            lg.warn('not enough identity sources, need to rotate again')
            self.automat('need-more-sources')
            return
        contacts_changed = False
        id_changed = my_id.rebuildLocalIdentity(
            new_sources=unique_sources,
            new_revision=self.new_revision,
        )
        new_contacts = my_id.getLocalIdentity().getContacts()
        if len(current_contacts) != len(new_contacts):
            contacts_changed = True
        if not contacts_changed:
            for pos in range(len(current_contacts)):
                if current_contacts[pos] != new_contacts[pos]:
                    contacts_changed = True
                    break
        self.rotated = True
        if _Debug:
            lg.args(_DebugLevel, new_sources=new_sources, contacts_changed=contacts_changed, id_changed=id_changed)
        self.automat('my-id-updated', (contacts_changed, id_changed))

    def doSendMyIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        d = self._do_send_my_identity()
        d.addCallback(lambda results: self.automat('my-id-sent', results))
        d.addErrback(lambda err: self.automat('my-id-failed', err))

    def doRequestMyIdentity(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'id_rotator.doRequestMyIdentity')

        def _cb(results):
            if _Debug:
                lg.args(_DebugLevel, results=results)
            # TODO: validate my identity in all responses
            self.automat('my-id-exist', results)
            return results

        def _eb(err):
            lg.err(err)
            self.automat('my-id-not-exist', err)
            if err.type == CancelledError:
                return None
            return err

        dl = []
        for idurl in my_id.getLocalIdentity().getSources(as_originals=True):
            dl.append(net_misc.getPageTwisted(idurl, timeout=15))
        d = DeferredList(dl, consumeErrors=True)
        d.addCallback(_cb)
        d.addErrback(_eb)

    def doReportDone(self, event, *args, **kwargs):
        """
        Action method.
        """
        if not self.result_defer:
            return
        if event == 'ping-done':
            self.result_defer.callback((self._is_healthy(args[0]), False))
        elif event == 'my-id-exist':
            self.result_defer.callback((
                True,
                self.rotated,
            ))
            # if self.rotated:
            #     events.send('my-identity-rotate-complete', data=dict())

    def doReportFailed(self, event, *args, **kwargs):
        """
        Action method.
        """
        result = args[0] if args else None
        lg.warn('id_rotator finished with failed result %r : %r' % (event, result))
        if not self.result_defer:
            return
        if event == 'auto-rotate-disabled':
            self.result_defer.errback(Exception('identity not healthy, but automatic rotate disabled'))
            return
        if event == 'no-id-servers-found':
            lg.warn('no more available identity servers found')
            self.result_defer.errback(Exception('no more available identity servers found'))
            return
        lg.err(result)
        self.result_defer.errback(result)

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.check_only = None
        self.result_defer = None
        self.alive_idurls = None
        self.current_servers = None
        self.preferred_servers = None
        self.known_servers = None
        self.possible_sources = None
        self.destroy()

    def _is_healthy(self, ping_results):
        # TODO: for now simply check that first identity server in my identity is alive
        # we can implement more "aggressive" checks later...
        # for example if any of my sources is not alive -> replace him also
        # also we can check here if I have enough sources according to those configs:
        #     services/identity-propagate/min-servers
        #     services/identity-propagate/max-servers
        if not ping_results:
            return False
        if not bool(ping_results[0]):
            return False
        unique_idurls = []
        for idurl in ping_results:
            if idurl not in unique_idurls:
                unique_idurls.append(idurl)
        min_servers = max(
            settings.MinimumIdentitySources(),
            config.conf().getInt('services/identity-propagate/min-servers') or settings.MinimumIdentitySources(),
        )
        max_servers = min(
            settings.MaximumIdentitySources(),
            config.conf().getInt('services/identity-propagate/max-servers') or settings.MaximumIdentitySources(),
        )
        if _Debug:
            lg.args(_DebugLevel, min_servers=min_servers, max_servers=max_servers, ping_results=ping_results, unique_idurls=unique_idurls)
        if len(unique_idurls) != len(ping_results):
            return False
        if len(unique_idurls) < min_servers:
            return False
        if len(unique_idurls) > max_servers:
            return False
        return True

    def _do_check_ping_results(self, ping_results):
        self.alive_idurls = []
        my_sources = my_id.getLocalIdentity().getSources(as_originals=True)
        local_revision = my_id.getLocalIdentity().getRevisionValue()
        latest_revision = -1
        pos = -1
        for result, remote_identity_src in ping_results:
            pos += 1
            idurl_bin = my_sources[pos]
            if not result:
                self.alive_idurls.append(None)
                continue
            try:
                remote_ident = identity.identity(xmlsrc=remote_identity_src)
                if not remote_ident.isCorrect():
                    raise Exception('remote identity not correct at position %r' % pos)
                if not remote_ident.Valid():
                    raise Exception('remote identity not valid at position %r' % pos)
                if latest_revision == -1:
                    latest_revision = remote_ident.getRevisionValue()
                if latest_revision <= remote_ident.getRevisionValue():
                    latest_revision = remote_ident.getRevisionValue()
            except:
                lg.exc(remote_identity_src)
                self.alive_idurls.append(None)
                continue
            if idurl_bin not in self.alive_idurls:
                self.alive_idurls.append(idurl_bin)
        if not self.new_revision:
            self.new_revision = max(local_revision, latest_revision) + 1
        if _Debug:
            lg.args(_DebugLevel, new_revision=self.new_revision, alive_idurls=self.alive_idurls)

        if not self.alive_idurls or not list(filter(None, self.alive_idurls)):
            # so all my id sources are down
            # if no alive sources found then probably network is down
            # and we should not do anything at the moment
            # but we must also check in that case if any of my ID servers are still alive, but only my identity was removed
            # otherwise we can get situation when all my id servers are UP, but my identity just expired and id_rotator do nothing to fix that
            self._fallback_and_ping_my_servers()
            return

        self.automat('ping-done', self.alive_idurls)

    def _do_send_my_identity(self):
        """
        Send my updated identity to the identity servers to register it.
        """
        my_sources = my_id.getLocalIdentity().getSources(as_originals=True)
        payload = my_id.getLocalIdentity().serialize(as_text=False)
        dlist = []
        if _Debug:
            lg.out(_DebugLevel, 'id_rotator._do_send_my_identity my_sources=%r' % my_sources)
        for idurl_bin in my_sources:
            _, host, _webport, filename = nameurl.UrlParse(idurl_bin)
            webport = None
            if host in self.preferred_servers:
                webport = int(self.preferred_servers[host][0])
            if not webport and host in self.known_servers:
                webport = int(self.known_servers[host][0])
            if not webport:
                webport = _webport
            url = net_misc.pack_address(
                (
                    host,
                    webport,
                ),
                proto='http',
            )
            dlist.append(net_misc.http_post_data(
                url=url,
                data=payload,
                connectTimeout=15,
            ))
            if _Debug:
                lg.args(_DebugLevel, url=url, filename=filename, size=len(payload))
        return DeferredList(dlist, fireOnOneCallback=True)

    def _do_verify_my_sources(self):
        my_sources = my_id.getLocalIdentity().getSources(as_originals=True)
        dl = []
        for idurl_bin in my_sources:
            d = net_misc.getPageTwisted(idurl_bin, timeout=15)
            dl.append(d)
        d = DeferredList(dl, consumeErrors=True)
        d.addCallback(self._do_check_ping_results)
        d.addErrback(lambda err: self.automat('ping-failed', err))

    def _fallback_and_ping_my_servers(self):
        """
        Just ping all my id servers by sending a HTTP request to the "main page".
        """
        my_sources = my_id.getLocalIdentity().getSources(as_originals=True)
        id_servers = []
        for idurl_bin in my_sources:
            proto, host, port, _ = nameurl.UrlParse(idurl_bin)
            id_servers.append(nameurl.UrlMake(proto, host, port, ''))
        if _Debug:
            lg.args(_DebugLevel, id_servers=id_servers)
        dl = []
        for url in id_servers:
            d = net_misc.getPageTwisted(url, timeout=15)
            dl.append(d)
        d = DeferredList(dl, consumeErrors=True)
        d.addCallback(lambda result: self.automat('ping-done', []))
        d.addErrback(self._fallback_check_network_connected)

    def _fallback_check_network_connected(self, err):
        if _Debug:
            lg.args(_DebugLevel, err)
        from bitdust.p2p import network_connector
        if network_connector.A().state != 'CONNECTED':
            self.automat('ping-failed', err)
        else:
            self.automat('ping-done', [])


#------------------------------------------------------------------------------


def main():
    from bitdust.system import bpio
    bpio.init()
    settings.init()
    lg.set_debug_level(20)
    from twisted.internet import reactor  # @UnresolvedImport
    ir = IdRotator()
    ir.addStateChangedCallback(lambda *a: reactor.stop(), oldstate=None, newstate='DONE')  # @UndefinedVariable
    ir.addStateChangedCallback(lambda *a: reactor.stop(), oldstate=None, newstate='FAILED')  # @UndefinedVariable
    reactor.callWhenRunning(ir.automat, 'run')  # @UndefinedVariable
    reactor.run()  # @UndefinedVariable
    settings.shutdown()


#------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#!/usr/bin/python
# online_status.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (online_status.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: online_status
.. role:: red

BitDust online_status() Automat

    <a href="https://bitdust.io/automats/online_status/online_status.png" target="_blank">
    <img src="https://bitdust.io/automats/online_status/online_status.png" style="max-width:100%;">
    </a>

A state machine and several extra methods to keep track of current users's online state.
To do p2p communications need to know who is available and who is not at the moment.

This simple state machine is used to detect connection status with remote user.
The situation when remote user replies to a packet sent to him
means that he is currently available over the network.

A one instance of ``online_status()`` machine is created for
every remote contact and monitor his status.


EVENTS:
    * :red:`ack-receieved`
    * :red:`ack-received`
    * :red:`handshake`
    * :red:`inbox-packet`
    * :red:`init`
    * :red:`offline-check`
    * :red:`ping-failed`
    * :red:`ping-now`
    * :red:`shook-up-hands`
    * :red:`shutdown`
    * :red:`timer-1min`
    * :red:`timer-20sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

from twisted.internet.task import LoopingCall
from twisted.internet.defer import Deferred

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import strng
from bitdust.lib import utime

from bitdust.contacts import contactsdb

from bitdust.main import events
from bitdust.main import listeners

from bitdust.p2p import ratings
from bitdust.p2p import commands
from bitdust.p2p import handshaker

from bitdust.transport import callback

from bitdust.userid import my_id
from bitdust.userid import id_url
from bitdust.userid import global_id

#------------------------------------------------------------------------------

_StatusLabels = {
    'CONNECTED': 'online',
    'PING?': 'checking',
    'OFFLINE': 'offline',
}

#------------------------------------------------------------------------------

_OnlineStatusDict = {}
_ShutdownFlag = False
_OfflineCheckTask = None

#------------------------------------------------------------------------------


def init():
    """
    Called from top level code when the software is starting.
    Needs to be called before other methods here.
    """
    global _OfflineCheckTask
    global _ShutdownFlag
    if _Debug:
        lg.out(_DebugLevel, 'online_status.init')
    _ShutdownFlag = False
    callback.insert_inbox_callback(1, Inbox)  # try to not overwrite top callback in the list, but stay on top
    callback.add_queue_item_status_callback(OutboxStatus)
    _OfflineCheckTask = LoopingCall(RunOfflineChecks)
    _OfflineCheckTask.start(10, now=False)


def shutdown():
    """
    Called from top level code when the software is stopping.
    """
    global _OfflineCheckTask
    global _ShutdownFlag
    global _OnlineStatusDict
    if _Debug:
        lg.out(_DebugLevel, 'online_status.shutdown')
    handshaker.cancel_all()
    _OfflineCheckTask.stop()
    del _OfflineCheckTask
    _OfflineCheckTask = None
    callback.remove_inbox_callback(Inbox)
    callback.remove_queue_item_status_callback(OutboxStatus)
    for o_status in list(_OnlineStatusDict.values()):
        o_status.automat('shutdown')
    _OnlineStatusDict.clear()
    _ShutdownFlag = True


#------------------------------------------------------------------------------


def online_statuses():
    global _OnlineStatusDict
    return _OnlineStatusDict


def check_create(idurl, keep_alive=True):
    """
    Creates new instance of online_status() state machine and send "init" event to it.
    """
    idurl = strng.to_bin(idurl)
    if id_url.is_empty(idurl):
        return False
    if not id_url.is_cached(idurl):
        return False
    idurl = id_url.field(idurl)
    if idurl not in list(_OnlineStatusDict.keys()):
        A(idurl, 'init', keep_alive=keep_alive)
        if _Debug:
            lg.out(_DebugLevel, 'online_status.check_create instance for %r was not found, made a new with state OFFLINE' % idurl)
    return True


#------------------------------------------------------------------------------


def on_ping_failed(err, idurl=None, channel=None):
    if _Debug:
        lg.args(_DebugLevel, err=err, idurl=idurl, channel=channel)
    return err


def ping(idurl, channel=None, ack_timeout=15, ping_retries=0, keep_alive=False):
    """
    Doing handshake with remote node only if it is currently not connected.
    Returns Deferred object.
    """
    idurl = strng.to_bin(idurl)
    if _Debug:
        lg.args(_DebugLevel, idurl=idurl, keep_alive=keep_alive, channel=channel)
    result = Deferred()
    result.addErrback(on_ping_failed, idurl=idurl, channel=channel)
    if id_url.is_empty(idurl):
        result.errback(Exception('empty idurl provided'))
        return result
    if not id_url.is_cached(idurl):
        if _Debug:
            lg.dbg(_DebugLevel, 'user identity %r not cached yet, executing clean handshake' % idurl)
        return handshaker.ping(
            idurl=idurl,
            ack_timeout=ack_timeout,
            ping_retries=ping_retries,
            channel=channel or 'clean_ping',
            keep_alive=keep_alive,
        )
    idurl = id_url.field(idurl)
    if not isKnown(idurl):
        if not check_create(idurl, keep_alive=keep_alive):
            raise Exception('can not create instance')
    A(idurl, 'ping-now', result, channel=channel, ack_timeout=ack_timeout, ping_retries=ping_retries, original_idurl=idurl.to_original())
    return result


def handshake(idurl, channel=None, ack_timeout=None, ping_retries=2, keep_alive=False):
    """
    Immediately doing handshake with remote node by fetching remote identity file and then
    sending my own Identity() to remote peer and wait for an Ack() packet.
    Returns Deferred object.
    """
    idurl = strng.to_bin(idurl)
    if _Debug:
        lg.args(_DebugLevel, idurl=idurl, keep_alive=keep_alive, channel=channel, ack_timeout=ack_timeout, ping_retries=ping_retries)
    result = Deferred()
    result.addErrback(on_ping_failed, idurl=idurl, channel=channel)
    if id_url.is_empty(idurl):
        result.errback(Exception('empty idurl provided'))
        return result
    if not id_url.is_cached(idurl):
        if _Debug:
            lg.dbg(_DebugLevel, 'user identity %r not cached yet, executing clean handshake' % idurl)
        return handshaker.ping(
            idurl=idurl,
            ack_timeout=ack_timeout,
            ping_retries=ping_retries,
            channel=channel or 'clean_handshake',
            keep_alive=keep_alive,
        )
    idurl = id_url.field(idurl)
    if not isKnown(idurl):
        if not check_create(idurl, keep_alive=keep_alive):
            raise Exception('can not create instance')
    A(idurl, 'handshake', result, channel=channel, ack_timeout=ack_timeout, ping_retries=ping_retries, original_idurl=idurl.to_original())
    return result


#------------------------------------------------------------------------------


def isKnown(idurl):
    """
    Return `True` if state machine online_status() already exists for this
    user.
    """
    global _OnlineStatusDict
    global _ShutdownFlag
    if _ShutdownFlag:
        return False
    if id_url.is_empty(idurl):
        return False
    if not id_url.is_cached(idurl):
        return False
    idurl = id_url.field(idurl)
    return idurl in list(_OnlineStatusDict.keys())


def isOnline(idurl):
    """
    Return True if given contact's state is ONLINE.
    """
    global _ShutdownFlag
    if _ShutdownFlag:
        return False
    if id_url.is_empty(idurl):
        return False
    if not id_url.is_cached(idurl):
        return False
    idurl = id_url.field(idurl)
    if not isKnown(idurl):
        return False
    return A(idurl).state in [
        'CONNECTED',
        'PING?',
    ]


def isOffline(idurl):
    """
    Return True if given contact's state is OFFLINE.
    """
    global _ShutdownFlag
    if _ShutdownFlag:
        return True
    if id_url.is_empty(idurl):
        return True
    if not id_url.is_cached(idurl):
        return True
    idurl = id_url.field(idurl)
    if not isKnown(idurl):
        return True
    return A(idurl).state == 'OFFLINE'


def isCheckingNow(idurl):
    """
    Return True if given contact's state is PING or ACK?.
    """
    global _ShutdownFlag
    if _ShutdownFlag:
        return False
    if id_url.is_empty(idurl):
        return False
    if not id_url.is_cached(idurl):
        return False
    idurl = id_url.field(idurl)
    if not isKnown(idurl):
        return False
    return A(idurl).state == 'PING'


def getInstance(idurl, autocreate=True):
    if _ShutdownFlag:
        return None
    if id_url.is_empty(idurl):
        return None
    if not id_url.is_cached(idurl):
        return None
    idurl = id_url.field(idurl)
    if not isKnown(idurl) and not autocreate:
        return None
    check_create(idurl)
    return A(idurl)


def stateToLabel(state, default='?'):
    global _StatusLabels
    return _StatusLabels.get(state, default)


def getCurrentState(idurl):
    """
    Return the current state of that user or `None` if that contact is unknown.
    """
    global _ShutdownFlag
    if _ShutdownFlag:
        return None
    if id_url.is_empty(idurl):
        return None
    if not id_url.is_cached(idurl):
        return None
    idurl = id_url.field(idurl)
    if not isKnown(idurl):
        return None
    return A(idurl).state


def getStatusLabel(idurl):
    """
    Return some text description about the current state of that user.
    """
    global _ShutdownFlag
    if _ShutdownFlag:
        return '?'
    if id_url.is_empty(idurl):
        return '?'
    if not id_url.is_cached(idurl):
        return '?'
    idurl = id_url.field(idurl)
    if not isKnown(idurl):
        return '?'
    return stateToLabel(A(idurl).state)


def listOfflineSuppliers(customer_idurl=None):
    """
    Loops all suppliers and check their state, return a list of those with
    state OFFLINE.
    """
    result = []
    for idurl in contactsdb.suppliers(customer_idurl=customer_idurl):
        if not idurl:
            result.append(idurl)
        elif isOffline(idurl):
            result.append(idurl)
    return result


def listOfflineCustomers():
    """
    Loops all customers and check their state, return a list of those with
    state OFFLINE.
    """
    result = []
    for idurl in contactsdb.customers():
        if not idurl:
            result.append(idurl)
        elif isOffline(idurl):
            result.append(idurl)
    return result


def countOfflineAmong(idurls_list):
    """
    Loops all IDs in ``idurls_list`` and count how many is OFFLINE.
    """
    num = 0
    for idurl in idurls_list:
        if isOffline(idurl):
            num += 1
    return num


def countOnlineAmong(idurls_list):
    """
    Loops all IDs in ``idurls_list`` and count how many is ONLINE.
    """
    num = 0
    for idurl in idurls_list:
        if idurl:
            if isOnline(idurl):
                num += 1
    return num


#------------------------------------------------------------------------------


def add_online_status_listener_callback(idurl, callback_method, oldstate=None, newstate=None, callback_id=None):
    """
    Attach new listener callback to the corresponding ``online_contact()`` automat for node with ``idurl``.
    """
    online_status_instance = getInstance(idurl)
    if not online_status_instance:
        return False
    online_status_instance.addStateChangedCallback(
        callback_method,
        newstate=newstate,
        oldstate=oldstate,
        callback_id=callback_id,
    )
    return True


def remove_online_status_listener_callback(idurl, callback_method=None, callback_id=None):
    """
    Release already added listener callback from the corresponding ``online_contact()`` automat.
    """
    online_status_instance = getInstance(idurl)
    if not online_status_instance:
        return False
    return online_status_instance.removeStateChangedCallback(cb=callback_method, callback_id=callback_id) > 0


#------------------------------------------------------------------------------


def populate_online_statuses():
    for online_s in online_statuses().values():
        listeners.push_snapshot('online_status', snap_id=online_s.idurl.to_text(), data=online_s.to_json())


#------------------------------------------------------------------------------


def OutboxStatus(pkt_out, status, error=''):
    """
    This method is called when raised a status report after sending a packet to
    remote peer.

    If packet sending was failed - user seems to be OFFLINE.
    """
    global _ShutdownFlag
    if _ShutdownFlag:
        return False
    if pkt_out.outpacket.RemoteID.to_bin() == my_id.getIDURL().to_bin():
        return False
    if pkt_out.outpacket.CreatorID.to_bin() != my_id.getIDURL().to_bin():
        return False
    if status == 'finished':
        if error == 'unanswered' and pkt_out.outpacket.Command == commands.Identity():
            if pkt_out.outpacket.OwnerID == my_id.getIDURL() and pkt_out.outpacket.CreatorID == my_id.getIDURL():
                # if not handshaker.is_running(pkt_out.outpacket.RemoteID):
                if _Debug:
                    lg.dbg(_DebugLevel, 'ping packet %s addressed to %r was "unanswered"' % (pkt_out, pkt_out.outpacket.RemoteID))
    else:
        if _Debug:
            lg.dbg(_DebugLevel, 'packet %s is "%s" with %s error: %r' % (pkt_out, status, pkt_out.outpacket, error))
    if pkt_out.outpacket.Command == commands.Identity():
        if pkt_out.outpacket.OwnerID == my_id.getIDURL() and pkt_out.outpacket.CreatorID == my_id.getIDURL():
            if handshaker.is_running(pkt_out.outpacket.RemoteID):
                handshaker.on_identity_packet_outbox_status(pkt_out, status, error)
    return False


def Inbox(newpacket, info, status, message):
    """
    This is called when some ``packet`` was received from remote peer - user seems to be ONLINE.
    """
    global _ShutdownFlag
    if _ShutdownFlag:
        return False
    if id_url.is_cached(newpacket.OwnerID):
        if newpacket.OwnerID == my_id.getIDURL():
            return False
    else:
        if newpacket.OwnerID.to_bin() == my_id.getIDURL().to_bin():
            return False
    if not id_url.is_cached(newpacket.OwnerID):
        return False
    if newpacket.RemoteID != my_id.getIDURL():
        return False
    check_create(newpacket.OwnerID)
    A(newpacket.OwnerID, 'inbox-packet', (newpacket, info, status, message))
    return False


def PacketSendingTimeout(remoteID, packetID):
    """
    Called from ``p2p.io_throttle`` when some packet is timed out.

    Right now this do nothing, to be improved.
    """
    global _ShutdownFlag
    if _ShutdownFlag:
        return False
    if remoteID == my_id.getIDURL():
        return False
    check_create(remoteID)
    # TODO: do something in that case ... send event "ping-now" ?
    A(remoteID, 'sent-timeout', packetID)
    return True


#------------------------------------------------------------------------------


def RunOfflineChecks():
    for o_status in list(_OnlineStatusDict.values()):
        if o_status.state != 'OFFLINE':
            # if user is online or checking: do nothing
            continue
        if not o_status.latest_check_time:
            # if no checks done yet but he is offline: ping user
            o_status.automat('offline-check')
            continue
        if utime.utcnow_to_sec1970() - o_status.latest_check_time > 10*60:
            # user is offline and latest check was sent a while ago: lets try to ping user again
            o_status.automat('offline-check')
            continue
        if o_status.latest_inbox_time and utime.utcnow_to_sec1970() - o_status.latest_inbox_time < 60:
            # user is offline, but we know that he was online recently: lets try to ping him again
            o_status.automat('offline-check')
            continue
    return True


#------------------------------------------------------------------------------


def A(idurl, event=None, *args, **kwargs):
    """
    Access method to interact with a state machine created for given contact.
    """
    global _ShutdownFlag
    global _OnlineStatusDict
    idurl = id_url.field(idurl)
    if idurl not in _OnlineStatusDict:
        if _ShutdownFlag:
            return None
        if not event:
            return None
        _OnlineStatusDict[idurl] = OnlineStatus(
            idurl=idurl,
            name='online_%s' % global_id.UrlToGlobalID(idurl),
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=False,
            log_transitions=_Debug,
        )
    if event is not None:
        _OnlineStatusDict[idurl].automat(event, *args, **kwargs)
    return _OnlineStatusDict[idurl]


#------------------------------------------------------------------------------


class OnlineStatus(automat.Automat):
    """
    This class implements all the functionality of ``online_status()`` state machine.
    """

    timers = {
        'timer-1min': (60, ['CONNECTED']),
        'timer-20sec': (20.0, ['PING?']),
    }

    def __init__(self, idurl, name, state, debug_level=0, log_events=False, log_transitions=False, **kwargs):
        """
        Builds `online_status()` state machine.
        """
        self.idurl = idurl
        self.latest_inbox_time = None
        self.latest_check_time = None
        self.keep_alive = False
        super(OnlineStatus, self).__init__(name=name, state=state, debug_level=debug_level, log_events=log_events, log_transitions=log_transitions, **kwargs)
        if _Debug:
            lg.out(_DebugLevel, 'online_status.ContactStatus %s %s %s' % (name, state, idurl))

    def to_json(self):
        j = super().to_json()
        glob_id = global_id.ParseIDURL(self.idurl)
        j.update({
            'idurl': self.idurl.to_text(),
            'global_id': glob_id['customer'],
            'idhost': glob_id['idhost'],
            'username': glob_id['user'],
        })
        return j

    def init(self):
        """
        Method to initialize additional variables and flags
        at creation phase of `online_status()` machine.
        """

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        Method to catch the moment when `online_status()` state were changed.
        """
        if _Debug:
            lg.out(_DebugLevel, '%s : [%s]->[%s]' % (self.name, oldstate, newstate))
        if newstate == 'CONNECTED':
            lg.info('remote node connected : %s' % self.idurl)
            events.send('node-connected', data=dict(
                global_id=self.idurl.to_id(),
                idurl=self.idurl,
                old_state=oldstate,
                new_state=newstate,
            ))
            listeners.push_snapshot('online_status', snap_id=self.idurl.to_text(), data=self.to_json())
        if newstate == 'OFFLINE' and oldstate != 'AT_STARTUP':
            lg.info('remote node disconnected : %s' % self.idurl)
            events.send('node-disconnected', data=dict(
                global_id=self.idurl.to_id(),
                idurl=self.idurl,
                old_state=oldstate,
                new_state=newstate,
            ))
            listeners.push_snapshot('online_status', snap_id=self.idurl.to_text(), data=self.to_json())
        if newstate == 'PING?' and oldstate != 'AT_STARTUP':
            listeners.push_snapshot('online_status', snap_id=self.idurl.to_text(), data=self.to_json())

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <http://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'OFFLINE'
                self.doInit(*args, **kwargs)
        #---PING?---
        elif self.state == 'PING?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doReportOffline(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'ping-failed' or event == 'timer-20sec':
                self.state = 'OFFLINE'
                self.doReportOffline(*args, **kwargs)
            elif event == 'handshake' or event == 'ping-now':
                self.doSetCallback(*args, **kwargs)
            elif event == 'inbox-packet' or event == 'ack-received' or event == 'shook-up-hands':
                self.state = 'CONNECTED'
                self.doRememberTime(*args, **kwargs)
                self.doReportConnected(*args, **kwargs)
        #---CONNECTED---
        elif self.state == 'CONNECTED':
            if event == 'ping-failed':
                self.state = 'OFFLINE'
                self.doReportOffline(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doReportOffline(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'timer-1min' and not self.isRecentInbox(*args, **kwargs):
                self.doHandshake(event, *args, **kwargs)
            elif event == 'handshake':
                self.state = 'PING?'
                self.doSetCallback(*args, **kwargs)
                self.doHandshake(event, *args, **kwargs)
            elif event == 'ping-now':
                self.doSetCallback(*args, **kwargs)
                self.doReportAlreadyConnected(*args, **kwargs)
            elif event == 'inbox-packet' or event == 'shook-up-hands' or event == 'ack-receieved':
                self.doRememberTime(*args, **kwargs)
        #---OFFLINE---
        elif self.state == 'OFFLINE':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'handshake' or event == 'ping-now':
                self.state = 'PING?'
                self.doSetCallback(*args, **kwargs)
                self.doHandshake(event, *args, **kwargs)
            elif event == 'inbox-packet' or event == 'shook-up-hands':
                self.state = 'CONNECTED'
                self.doRememberTime(*args, **kwargs)
                self.doReportConnected(*args, **kwargs)
            elif event == 'offline-check' or event == 'ack-received':
                self.doRememberCheckTime(*args, **kwargs)
                self.doHandshake(event, *args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def isRecentInbox(self, *args, **kwargs):
        """
        Condition method.
        """
        if handshaker.is_running(self.idurl.to_bin()):
            return True
        if not self.latest_inbox_time:
            return False
        return utime.utcnow_to_sec1970() - self.latest_inbox_time > 60

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.handshake_callbacks = []
        self.keep_alive = kwargs.get('keep_alive', True)

    def doSetCallback(self, *args, **kwargs):
        """
        Action method.
        """
        if args and args[0]:
            self.handshake_callbacks.append(args[0])

    def doHandshake(self, event, *args, **kwargs):
        """
        Action method.
        """
        channel = kwargs.get('channel', None)
        ack_timeout = kwargs.get('ack_timeout', 15)
        ping_retries = kwargs.get('ping_retries', 2)
        original_idurl = kwargs.get('original_idurl', self.idurl.to_bin())
        d = None
        if event == 'ping-now':
            d = handshaker.ping(
                idurl=original_idurl,
                ack_timeout=ack_timeout,
                ping_retries=ping_retries,
                channel=channel or 'ping',
                cancel_running=True,
            )
        elif event == 'handshake':
            d = handshaker.ping(
                idurl=original_idurl,
                ack_timeout=ack_timeout,
                ping_retries=ping_retries,
                force_cache=True,
                channel=channel or 'handshake',
                cancel_running=True,
            )
        elif event == 'offline-ping':
            if self.keep_alive:
                d = handshaker.ping(
                    idurl=original_idurl,
                    ack_timeout=ack_timeout,
                    cache_timeout=15,
                    ping_retries=ping_retries,
                    force_cache=True,
                    channel='offline_ping',
                )
        else:
            if self.keep_alive:
                d = handshaker.ping(
                    idurl=original_idurl,
                    ack_timeout=ack_timeout,
                    cache_timeout=15,
                    ping_retries=ping_retries,
                    force_cache=True,
                    channel='idle_ping',
                )
        if d:
            d.addCallback(self._on_ping_success)
            d.addErrback(self._on_ping_failed)

    def doRememberTime(self, *args, **kwargs):
        """
        Action method.
        """
        to_be_remembered = True
        if self.latest_inbox_time:
            if utime.utcnow_to_sec1970() - self.latest_inbox_time < 5*60:
                to_be_remembered = False
        if to_be_remembered:
            ratings.remember_connected_time(self.idurl.to_bin())
        self.latest_inbox_time = utime.utcnow_to_sec1970()

    def doRememberCheckTime(self, *args, **kwargs):
        """
        Action method.
        """
        self.latest_check_time = utime.utcnow_to_sec1970()

    def doReportAlreadyConnected(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.args(_DebugLevel, idurl=self.idurl, keep_alive=self.keep_alive, handshake_callbacks=len(self.handshake_callbacks))
        for cb in self.handshake_callbacks:
            if isinstance(cb, Deferred):
                if not cb.called:
                    cb.callback(None)
            else:
                cb(None)
        self.handshake_callbacks = []

    def doReportOffline(self, *args, **kwargs):
        """
        Action method.
        """
        err = args[0] if (args and args[0]) else Exception('user is offline')
        try:
            err_msg = err.getErrorMessage()
        except:
            err_msg = repr(err)
        if _Debug:
            lg.args(_DebugLevel, idurl=self.idurl, err=err_msg, keep_alive=self.keep_alive, handshake_callbacks=len(self.handshake_callbacks))
        for cb in self.handshake_callbacks:
            if isinstance(cb, Deferred):
                if not cb.called:
                    cb.errback(err)
            else:
                cb(err)
        self.handshake_callbacks = []

    def doReportConnected(self, *args, **kwargs):
        """
        Action method.
        """
        response = args[0] if (args and args[0]) else None
        if _Debug:
            lg.args(_DebugLevel, idurl=self.idurl, response=response, keep_alive=self.keep_alive, handshake_callbacks=len(self.handshake_callbacks))
        for cb in self.handshake_callbacks:
            if isinstance(cb, Deferred):
                if not cb.called:
                    cb.callback(response)
            else:
                cb(response)
        self.handshake_callbacks = []

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        global _OnlineStatusDict
        _OnlineStatusDict.pop(self.idurl)
        self.idurl = None
        self.latest_inbox_time = None
        self.handshake_callbacks = None
        self.destroy()

    def _on_ping_success(self, result):
        try:
            response = result[0]
            info = result[1]
        except:
            lg.exc()
        if _Debug:
            lg.out(_DebugLevel, 'online_status._on_ping_success %r : %r' % (self.idurl, result))
        self.automat('shook-up-hands', (
            response,
            info,
        ))
        return None

    def _on_ping_failed(self, err):
        try:
            msg = err.getErrorMessage()
        except:
            msg = str(err)
        if _Debug:
            lg.out(_DebugLevel, 'online_status._on_ping_failed %r : %s' % (self.idurl, msg))
        self.automat('ping-failed', err)
        return None



#!/usr/bin/env python3

import socks, connections, time, sys, json
import options

config = options.Get()
config.read()
version = config.version

# print ('Number of arguments:', len(sys.argv), 'arguments.')
# print ('Argument List:', str(sys.argv))

try:
    command = sys.argv[1]

    try:
        arg1 = sys.argv[2]
    except:
        pass

    try:
        arg2 = sys.argv[3]
    except:
        pass

    try:
        arg3 = sys.argv[4]
    except:
        pass

    try:
        arg4 = sys.argv[5]
    except:
        pass

    try:
        arg5 = sys.argv[6]
    except:
        pass

except:

    entry = input('No argument detected, please insert command manually\n').split()
    command = entry[0]
    print(entry)

    try:
        arg1 = entry[1]
    except:
        pass
    try:
        arg2 = entry[2]
    except:
        pass
    try:
        arg3 = entry[3]
    except:
        pass
    try:
        arg4 = entry[4]
    except:
        pass
    try:
        arg5 = entry[5]
    except:
        pass

    try:
        arg6 = entry[6]
    except:
        pass

s = socks.socksocket()
s.settimeout(10)

is_regnet = False

if 'testnet' in version:
    s.connect(('127.0.0.1', 2829))
    print('tesnet mode')
elif 'regnet' in version:
    is_regnet = True
    print('Regtest mode')
    s.connect(('127.0.0.1', 3030))
else:
    s.connect(('127.0.0.1', 5658))
    #s.connect(("34.192.6.105", 5658))
    #s.connect(("bismuth.live", 5658))


def api_getconfig(socket):
    connections.send(s, 'api_getconfig')
    result = connections.receive(s)
    print(result)


def diffget(socket):
    #check difficulty
    connections.send(s, 'diffget')
    diff = connections.receive(s)
    print('Current difficulty: {}'.format(diff))
    #check difficulty


def diffgetjson(socket):
    #check difficulty
    connections.send(s, 'diffgetjson')
    response = connections.receive(s)
    #for key in response:
    #    print (key,":",response[key])
    print(json.dumps(response))
    #check difficulty


def balanceget(socket, arg1):
    #get balance
    connections.send(s, 'balanceget')
    connections.send(s, arg1)
    balanceget_result = connections.receive(s)
    print('Address balance: {}'.format(balanceget_result[0]))
    print('Address credit: {}'.format(balanceget_result[1]))
    print('Address debit: {}'.format(balanceget_result[2]))
    print('Address fees: {}'.format(balanceget_result[3]))
    print('Address rewards: {}'.format(balanceget_result[4]))
    print('Address balance without mempool: {}'.format(balanceget_result[5]))
    #get balance


def balancegetjson(socket, arg1):
    #get balance
    connections.send(s, 'balancegetjson')
    connections.send(s, arg1)
    response = connections.receive(s)
    print(json.dumps(response))
    #get balance


def balancegethyper(socket, arg1):
    #get balance
    connections.send(s, 'balancegethyper')
    connections.send(s, arg1)
    balanceget_result = connections.receive(s)
    print('Address balance: {}'.format(balanceget_result))
    #get balance


def balancegethyperjson(socket, arg1):
    #get balance
    connections.send(s, 'balancegethyperjson')
    connections.send(s, arg1)
    response = connections.receive(s)
    print(json.dumps(response))
    #get balance


#insert to mempool
#DIRECT INSERT, NO REMOTE TX CONSTRUCTION
def mpinsert(s, transaction):
    connections.send(s, 'mpinsert')
    connections.send(s, transaction)
    confirmation = connections.receive(s)
    print(confirmation)


#insert to mempool


def mpget(socket):
    #ask for mempool
    connections.send(s, 'mpget')
    mempool = connections.receive(s)
    print('Current mempool: {}'.format(mempool))
    #ask for mempool


def mpgetjson(socket):
    #ask for mempool
    connections.send(s, 'mpgetjson')
    response_list = connections.receive(s)
    print('Current mempool:')
    print(json.dumps(response_list))
    #ask for mempool


def difflast(socket):
    #ask for last difficulty
    connections.send(s, 'difflast')
    response = connections.receive(s)
    blocklast = response[0]
    difflast = response[1]
    print('Last block: {}'.format(blocklast))
    print('Last difficulty: {}'.format(difflast))
    #ask for last difficulty


def difflastjson(socket):
    #ask for last difficulty
    connections.send(s, 'difflastjson')
    response = connections.receive(s)
    print(json.dumps(response))
    #ask for last difficulty


def blocklast(socket):
    #get last block
    connections.send(s, 'blocklast')
    block_last = connections.receive(s)

    print('Last block number: {}'.format(block_last[0]))
    print('Last block timestamp: {}'.format(block_last[1]))
    print('Last block hash: {}'.format(block_last[7]))
    #get last block


def blocklastjson(socket):
    #get last block
    connections.send(s, 'blocklastjson')
    response = connections.receive(s)
    print(json.dumps(response))
    #get last block


def api_getblocksince(socket, arg1=None):
    #get last block
    connections.send(s, 'api_getblocksince')
    if arg1:
        connections.send(s, arg1)
    response = connections.receive(s)
    print(json.dumps(response))
    #get last block


def keygen(socket):
    #generate address
    #RECEIVES PRIVATE KEY FROM NODE
    connections.send(s, 'keygen')
    keys_generated = connections.receive(s)

    print('Private key: {}'.format(keys_generated[0]))
    print('Public key: {}'.format(keys_generated[1]))
    print('Address: {}'.format(keys_generated[2]))
    #generate address


def keygenjson(socket):
    #generate address
    #RECEIVES PRIVATE KEY FROM NODE
    connections.send(s, 'keygenjson')
    response = connections.receive(s)
    print(json.dumps(response))
    #generate address


def blockget(socket, arg1):
    #get block
    connections.send(s, 'blockget')
    connections.send(s, arg1)
    block_get = connections.receive(s)
    print('Requested block: {}'.format(block_get))
    print('Requested block number of transactions: {}'.format(len(block_get)))
    print('Requested block height: {}'.format(block_get[0][0]))
    #get block


def blockgetjson(socket, arg1):
    #get block
    connections.send(s, 'blockgetjson')
    connections.send(s, arg1)
    response_list = connections.receive(s)
    print(json.dumps(response_list))
    #get block


def addlist(socket, arg1):
    #get all txs for an address
    connections.send(s, 'addlist')
    connections.send(s, arg1)
    address_tx_list = connections.receive(s)
    print('All transactions for requested address:')
    for row in address_tx_list:
        print(row)
    #get all txs for an address


def addlistlim(socket, arg1, arg2):
    #get x txs for an address
    connections.send(s, 'addlistlim')
    connections.send(s, arg1)
    connections.send(s, arg2)
    address_tx_list = connections.receive(s)
    print('Transactions for requested address:')
    for row in address_tx_list:
        print(row)
    #get all txs for an address


def addlistlimjson(socket, arg1, arg2):
    #get x txs for an address
    connections.send(s, 'addlistlimjson')
    connections.send(s, arg1)
    connections.send(s, arg2)
    response_list = connections.receive(s)
    print('Transactions for requested address:')
    print(json.dumps(response_list))
    #get all txs for an address


def addlistlimmir(socket, arg1, arg2):
    #get x negative txs for an address
    connections.send(s, 'addlistlimmir')
    connections.send(s, arg1)
    connections.send(s, arg2)
    address_tx_list = connections.receive(s)
    print('Mirror transactions for requested address:')
    for row in address_tx_list:
        print(row)
    #get all txs for an address


def addlistlimmirjson(socket, arg1, arg2):
    #get x negative txs for an address
    connections.send(s, 'addlistlimmirjson')
    connections.send(s, arg1)
    connections.send(s, arg2)
    response_list = connections.receive(s)
    print('Mirror transactions for requested address:')
    print(json.dumps(response_list))
    #get all txs for an address


def listlim(socket, arg1):
    #get x last txs
    connections.send(s, 'listlim')
    connections.send(s, arg1)
    tx_list = connections.receive(s)
    print('All transactions for requested range:')
    for row in tx_list:
        print(row)


def api_getblockfromhash(socket, arg1):
    connections.send(s, 'api_getblockfromhash')
    connections.send(s, arg1)
    reply = connections.receive(s)
    print(reply)


def listlimjson(socket, arg1):
    #get x last txs
    connections.send(s, 'listlimjson')
    connections.send(s, arg1)
    response_list = connections.receive(s)
    print('All transactions for requested range:')
    print(json.dumps(response_list))


def txsend(socket, arg1, arg2, arg3, arg4, arg5):
    #generate transaction
    #SENDS PRIVATE KEY TO NODE
    connections.send(s, 'txsend')

    remote_tx_timestamp = '%.2f' % time.time()
    remote_tx_privkey = arg1  # node will dump pubkey+address from this
    remote_tx_recipient = arg2
    remote_tx_amount = arg3
    remote_tx_operation = arg4
    remote_tx_openfield = arg5

    #connections.send(s, (remote_tx_timestamp, remote_tx_privkey, remote_tx_recipient, remote_tx_amount, remote_tx_keep, remote_tx_openfield))
    connections.send(s, (str(remote_tx_timestamp), str(remote_tx_privkey), str(remote_tx_recipient), str(remote_tx_amount), str(remote_tx_operation), str(remote_tx_openfield)))
    #generate transaction

    signature = connections.receive(s)
    print(signature)


def aliasget(socket, arg1):
    connections.send(s, 'aliasget')
    connections.send(s, arg1)
    alias_results = connections.receive(s)
    print(alias_results)


def tokensget(socket, arg1):
    connections.send(s, 'tokensget')
    connections.send(s, arg1)
    tokens_results = connections.receive(s)
    print(tokens_results)


def addfromalias(socket, arg1):
    connections.send(s, 'addfromalias')
    connections.send(s, arg1)
    address_fetch = connections.receive(s)
    print(address_fetch)


def peersget(socket):
    connections.send(s, 'peersget')
    peers_received = connections.receive(s)
    print(peers_received)


def statusget(socket):
    connections.send(s, 'statusjson')
    response = connections.receive(s)
    print(json.dumps(response))


def portget(socket):
    connections.send(s, 'portget')
    response = connections.receive(s)
    print(json.dumps(response))


def addvalidate(socket, arg1):
    connections.send(s, 'addvalidate')
    connections.send(s, arg1)
    validate_result = connections.receive(s)
    print(validate_result)


def aliasesget(socket, arg1):
    arg_split = arg1.split(',')
    print(arg_split)

    connections.send(s, 'aliasesget')
    connections.send(s, arg_split)
    alias_results = connections.receive(s)
    print(alias_results)


def api_getaddresssince(socket, arg1, arg2, arg3):
    connections.send(s, 'api_getaddresssince')
    connections.send(s, arg1)
    connections.send(s, arg2)
    connections.send(s, arg3)
    response = connections.receive(s)
    print(json.dumps(response))


if command == 'getversion':
    connections.send(s, 'getversion')
    print(connections.receive(s))

if command == 'generate':
    if not is_regnet:
        print('Only available on regnet')
        sys.exit()
    connections.send(s, 'regtest_generate')
    connections.send(s, arg1)
    print(connections.receive(s))

if command == 'mpfill':
    if not is_regnet:
        print('Only available on regnet')
        sys.exit()
    connections.send(s, 'regtest_mpfill')
    connections.send(s, arg1)
    print(connections.receive(s))

if command == 'mpinsert':
    #arg1 = '1520788207.69', '4edadac9093d9326ee4b17f869b14f1a2534f96f9c5d7b48dc9acaed', '4edadac9093d9326ee4b17f869b14f1a2534f96f9c5d7b48dc9acaed', '0.00000000', 'e0piKXvc636t0fYmxdOti3fJZ+G1vQYAJ2IZv4inPGQYgG4nS0lU+61LDQQVqeGvmsDOsxFhM6VVLpYExPmc5HF6e1ZAr5IXQ69s88sJBx/XVl1YavAdo0katGDyvZpQf609F8PVbtD0zzBinQjfkoXU/NXo00CEyniyYPxAXuI=', 'LS0tLS1CRUdJTiBQVUJMSUMgS0VZLS0tLS0KTUlHZk1BMEdDU3FHU0liM0RRRUJBUVVBQTRHTkFEQ0JpUUtCZ1FES3ZMVGJEeDg1YTF1Z2IvNnhNTWhWT3E2VQoyR2VZVDgrSXEyejlGd0lNUjQwbDJ0dEdxTks3dmFyTmNjRkxJdThLbjRvZ0RRczNXU1dRQ3hOa2haaC9GcXpGCllZYTMvSXRQUGZ6clhxZ2Fqd0Q4cTRadDRZbWp0OCsyQmtJbVBqakZOa3VUUUl6Mkl1M3lGcU9JeExkak13N24KVVZ1OXRGUGlVa0QwVm5EUExRSURBUUFCCi0tLS0tRU5EIFBVQkxJQyBLRVktLS0tLQ==', '0', ''
    mpinsert(s, arg1)

if command == 'aliasget':
    aliasget(s, arg1)

if command == 'tokensget':
    tokensget(s, arg1)

if command == 'addvalidate':
    addvalidate(s, arg1)

if command == 'aliasesget':
    aliasesget(s, arg1)

elif command == 'diffget':
    diffget(s)

elif command == 'diffgetjson':
    diffgetjson(s)

elif command == 'difflast':
    difflast(s)

elif command == 'difflastjson':
    difflastjson(s)

elif command == 'balanceget':
    balanceget(s, arg1)

elif command == 'balancegetjson':
    balancegetjson(s, arg1)

elif command == 'balancegethyper':
    balancegethyper(s, arg1)

elif command == 'balancegethyperjson':
    balancegethyperjson(s, arg1)

elif command == 'api_getconfig':
    api_getconfig(s)

elif command == 'mpget':
    mpget(s)

elif command == 'mpgetjson':
    mpgetjson(s)

elif command == 'statusget':
    statusget(s)

elif command == 'portget':
    portget(s)

elif command == 'peersget':
    peersget(s)

elif command == 'blocklast':
    blocklast(s)

elif command == 'blocklastjson':
    blocklastjson(s)

elif command == 'keygen':
    keygen(s)

elif command == 'keygenjson':
    keygenjson(s)

elif command == 'blockget':
    blockget(s, arg1)

elif command == 'blockgetjson':
    blockgetjson(s, arg1)

elif command == 'addlist':
    addlist(s, arg1)

elif command == 'addlistlim':
    addlistlim(s, arg1, arg2)

elif command == 'addlistlimjson':
    addlistlimjson(s, arg1, arg2)

elif command == 'addlistlimmir':
    addlistlimmir(s, arg1, arg2)

elif command == 'addlistlimmirjson':
    addlistlimmirjson(s, arg1, arg2)

elif command == 'listlim':
    listlim(s, arg1)

elif command == 'api_getblockfromhash':
    api_getblockfromhash(s, arg1)

elif command == 'api_getblocksince':
    try:
        api_getblocksince(s, int(arg1))
    except:
        api_getblocksince(s)

elif command == 'stop':
    connections.send(s, 'stop')
    print('Asked to Stop')

elif command == 'addfromalias':
    addfromalias(s, arg1)

elif command == 'api_getaddresssince':
    api_getaddresssince(s, arg1, arg2, arg3)

elif command == 'txsend':
    try:
        arg4
    except:
        arg4 = '0'

    try:
        arg5
    except:
        arg5 = ''

    txsend(s, arg1, arg2, arg3, arg4, arg5)

s.close()



#!/usr/bin/env python
# nickname_observer.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (nickname_observer.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: nickname_observer.

.. role:: red

BitDust nickname_observer() Automat

.. raw:: html

    <a href="nickname_observer.png" target="_blank">
    <img src="nickname_observer.png" style="max-width:100%;">
    </a>

EVENTS:
    * :red:`dht-read-failed`
    * :red:`dht-read-success`
    * :red:`find-one`
    * :red:`observe-many`
    * :red:`stop`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import sys

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.main import settings

from bitdust.userid import my_id

from bitdust.dht import dht_service
from bitdust.dht import dht_records

#------------------------------------------------------------------------------


def find_one(nickname, attempts=3, results_callback=None):
    if _Debug:
        lg.out(_DebugLevel, 'nickname_observer.find_one %s %d' % (nickname, attempts))
    observer = NicknameObserver(
        name='nickname_observer_%s' % nickname,
        state='AT_STARTUP',
        debug_level=_DebugLevel,
        log_events=_Debug,
        log_transitions=_Debug,
    )
    observer.automat('find-one', (nickname, attempts, results_callback))
    return observer


def observe_many(nickname, attempts=10, results_callback=None):
    if _Debug:
        lg.out(_DebugLevel, 'nickname_observer.observe_many %s %d' % (nickname, attempts))
    observer = NicknameObserver(
        name='nickname_observer_%s' % nickname,
        state='AT_STARTUP',
        debug_level=_DebugLevel,
        log_events=_Debug,
        log_transitions=_Debug,
    )
    observer.automat('observe-many', (nickname, attempts, results_callback))
    return observer


def stop_all():
    for a in automat.objects().values():
        if isinstance(a, NicknameObserver):
            if _Debug:
                lg.out(_DebugLevel, 'nickname_observer.stop_all sends "stop" to %r' % a)
            a.automat('stop')


#------------------------------------------------------------------------------


class NicknameObserver(automat.Automat):
    """
    This class implements all the functionality of the ``nickname_observer()``
    state machine.
    """
    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """
        self.nickname = None
        self.attempts = None
        self.key = None
        self.dht_read_defer = None
        self.result_callback = None

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'observe-many':
                self.state = 'DHT_LOOP'
                self.doInit(*args, **kwargs)
                self.doDHTReadKey(*args, **kwargs)
            elif event == 'find-one':
                self.state = 'DHT_FIND'
                self.doInit(*args, **kwargs)
                self.doDHTReadKey(*args, **kwargs)
        #---DHT_LOOP---
        elif self.state == 'DHT_LOOP':
            if event == 'stop':
                self.state = 'STOPPED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'dht-read-failed' and self.isMoreAttemptsNeeded(*args, **kwargs):
                self.doNextKey(*args, **kwargs)
                self.doDHTReadKey(*args, **kwargs)
            elif event == 'dht-read-success' and not self.isMoreAttemptsNeeded(*args, **kwargs):
                self.state = 'FINISHED'
                self.doReportNicknameExist(*args, **kwargs)
                self.doReportFinished(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'dht-read-success' and self.isMoreAttemptsNeeded(*args, **kwargs):
                self.doReportNicknameExist(*args, **kwargs)
                self.doNextKey(*args, **kwargs)
                self.doDHTReadKey(*args, **kwargs)
            elif event == 'dht-read-failed' and not self.isMoreAttemptsNeeded(*args, **kwargs):
                self.state = 'FINISHED'
                self.doReportFinished(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---DHT_FIND---
        elif self.state == 'DHT_FIND':
            if event == 'dht-read-success':
                self.state = 'FOUND'
                self.doReportNicknameExist(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'dht-read-failed' and self.isMoreAttemptsNeeded(*args, **kwargs):
                self.doNextKey(*args, **kwargs)
                self.doDHTReadKey(*args, **kwargs)
            elif event == 'stop':
                self.state = 'STOPPED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'dht-read-failed' and not self.isMoreAttemptsNeeded(*args, **kwargs):
                self.state = 'NOT_FOUND'
                self.doReportNicknameNotExist(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---FOUND---
        elif self.state == 'FOUND':
            pass
        #---STOPPED---
        elif self.state == 'STOPPED':
            pass
        #---NOT_FOUND---
        elif self.state == 'NOT_FOUND':
            pass
        #---FINISHED---
        elif self.state == 'FINISHED':
            pass
        return None

    def isMoreAttemptsNeeded(self, *args, **kwargs):
        """
        Condition method.
        """
        return self.attempts > 1

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.nickname, self.attempts, self.result_callback = args[0]
        try:
            nick, index = self.nickname.rsplit(':', 1)
            index = int(index)
        except:
            nick = self.nickname.replace(':', '_')
            index = 0
        self.nickname = nick
        # self.key = nick + ':' + str(number)
        self.key = dht_service.make_key(
            key=self.nickname,
            index=index,
            prefix='nickname',
        )

    def doNextKey(self, *args, **kwargs):
        """
        Action method.
        """
        try:
            key_info = dht_service.split_key(self.key)
            # nik, number = self.key.rsplit(':', 1)
            index = int(key_info['index'])
            # number = int(number)
        except:
            lg.exc()
            index = 0
        index += 1
        # self.key = self.nickname + ':' + str(number)
        self.key = dht_service.make_key(
            key=self.nickname,
            index=index,
            prefix='nickname',
        )
        self.attempts -= 1

    def doDHTReadKey(self, *args, **kwargs):
        """
        Action method.
        """
        if self.dht_read_defer is not None:
            self.dht_read_defer.cancel()
            self.dht_read_defer = None
        self.dht_read_defer = dht_records.get_nickname(self.key)
        self.dht_read_defer.addCallback(self._dht_read_result, self.key)
        self.dht_read_defer.addErrback(self._dht_read_failed)

    def doReportNicknameExist(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'nickname_observer.doReportNicknameExist : (%s, %s)' % (self.key, args[0]))
        if self.result_callback is not None:
            try:
                key_info = dht_service.split_key(self.key)
                nick = key_info['key']
                index = key_info['index']
            except:
                lg.exc()
                nick = self.nickname
                index = 0
            # nik, num = self.key.split(':')
            # num = int(num)
            self.result_callback('exist', nick, index, *args, **kwargs)

    def doReportNicknameNotExist(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'nickname_observer.doReportNicknameNotExist : %s' % self.nickname)
        if self.result_callback is not None:
            self.result_callback('not exist', self.nickname, -1, '')

    def doReportFinished(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'nickname_observer.doReportFinished')
        if self.result_callback is not None:
            self.result_callback('finished', '', -1, '')

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        if self.dht_read_defer is not None:
            self.dht_read_defer.pause()
            self.dht_read_defer.cancel()
            self.dht_read_defer = None
        self.result_callback = None
        self.destroy()

    def _dht_read_result(self, value, key):
        if self.dht_read_defer is None:
            return
        self.dht_read_defer = None
        if not value:
            self.automat('dht-read-failed')
            return
        try:
            v = value['idurl']
        except:
            lg.out(14, '%r' % value)
            lg.exc()
            self.automat('dht-read-failed')
            return
        self.automat('dht-read-success', v)

    def _dht_read_failed(self, x):
        if self.dht_read_defer is None:
            return
        self.automat('dht-read-failed', x)
        self.dht_read_defer = None


#------------------------------------------------------------------------------


def main():
    if len(sys.argv) < 3:
        print('usage: nickname_observer.py <"many"|"one"> <nickname> <attempts>')
        return
    from twisted.internet import reactor  # @UnresolvedImport
    lg.set_debug_level(24)
    settings.init()
    my_id.init()
    dht_service.init(settings.getDHTPort())

    def _result(result, nickname):
        print(result, nickname)
        if result == 'finished':
            reactor.stop()  # @UndefinedVariable

    if sys.argv[1] == 'many':
        observe_many(sys.argv[2], int(sys.argv[3]), results_callback=_result)
    else:
        find_one(sys.argv[2], int(sys.argv[3]), _result)
    reactor.run()  # @UndefinedVariable
    settings.shutdown()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# known_servers.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (known_servers.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_KnownServers = None

#------------------------------------------------------------------------------


def default_nodes():
    """
    A set of identity servers currently maintained, see file `default_network.json` in the root folder.
    """
    from bitdust.lib import strng
    from bitdust.main import network_config
    network_info = network_config.read_network_config_file()
    identity_servers = {}
    for identity_server in network_info['service_identity_propagate']['known_servers']:
        identity_servers[strng.to_bin(identity_server['host'])] = (
            identity_server['http_port'],
            identity_server.get('tcp_port', 6661),
        )
    return identity_servers


def by_host():
    """
    Here is a "well known" identity servers to support the network.
    Keys are domain names or global IP address (not recommended) of the ID server.
    Values are ``Web port`` (reading) and ``TCP port`` (writing) numbers.

    This is some kind of "genesis" network.
    If you willing to support the project and started your own BitDust node on reliable machine,
    contact us and we will include your address here.
    So other nodes will be able to use your machine to host their identities.

    You can override those "genesis" nodes by configuring list of your preferred identity servers
    in the program settings:

        api.config_set(
            "services/identity-propagate/known-servers",
            "myfirstserver.net:80, secondmachine.net:8080, thirdnode.gov.eu:80",
        )

    Also check file `default_network.json` in the repository root.

    Both ways you can create your own BitDust network, under your full control.
    """
    global _KnownServers
    from bitdust.main import config
    from bitdust.lib import strng

    if _KnownServers is not None:
        return _KnownServers

    try:
        overridden_identity_servers_str = str(config.conf().getString('services/identity-propagate/known-servers'))
    except:
        overridden_identity_servers_str = ''

    if not overridden_identity_servers_str:
        _KnownServers = default_nodes()
        return _KnownServers

    overridden_identity_servers = {}
    for id_server_str in overridden_identity_servers_str.split(','):
        if id_server_str.strip():
            try:
                id_server = id_server_str.strip().split(':')
                id_server_host = strng.to_bin(id_server[0].strip())
                id_server_web_port = int(id_server[1].strip())
                if len(id_server) > 2:
                    id_server_tcp_port = int(id_server[2].strip())
                else:
                    id_server_tcp_port = 6661
            except:
                continue
            overridden_identity_servers[id_server_host] = (
                id_server_web_port,
                id_server_tcp_port,
            )

    if overridden_identity_servers:
        _KnownServers = overridden_identity_servers
        return _KnownServers

    _KnownServers = default_nodes()
    return _KnownServers



#!/usr/bin/python
# diskspace.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (diskspace.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: diskspace.

This is set of methods to operate with amount of space units.
Examples::
    123 bytes
    45 KB
    67 mb
    8.9 Gigabytes

You can translate values from Kb to Mb or create a good looking string from bytes number.
"""

#------------------------------------------------------------------------------

_Suffixes = {
    '': 1.0,
    'bytes': 1.0,
    'b': 1.0,
    'B': 1.0,
    'K': 1024.0,
    'k': 1024.0,
    'KB': 1024.0,
    'Kb': 1024.0,
    'kb': 1024.0,
    'Kilobytes': 1024.0,
    'kilobytes': 1024.0,
    'm': 1024.0*1024.0,
    'M': 1024.0*1024.0,
    'MB': 1024.0*1024.0,
    'Mb': 1024.0*1024.0,
    'mb': 1024.0*1024.0,
    'Megabytes': 1024.0*1024.0,
    'megabytes': 1024.0*1024.0,
    'G': 1024.0*1024.0*1024.0,
    'GB': 1024.0*1024.0*1024.0,
    'Gb': 1024.0*1024.0*1024.0,
    'g': 1024.0*1024.0*1024.0,
    'gb': 1024.0*1024.0*1024.0,
    'Gigabytes': 1024.0*1024.0*1024.0,
    'gigabytes': 1024.0*1024.0*1024.0,
}

_SuffixLabels = ['bytes', 'Kb', 'Mb', 'Gb']

_MultiDict = {
    'Kb': 1024.0,
    'Mb': 1024.0*1024.0,
    'Gb': 1024.0*1024.0*1024.0,
}

#------------------------------------------------------------------------------


class DiskSpace:
    """
    You can create an object of this class and use it as variable to store
    amount of space.

    But in most cases this class is not used, see below global methods
    in this module.
    """
    def __init__(self, v=None, s='0Mb'):
        if v is None:
            self.v = s
            if self.getSZ() not in ['Kb', 'Mb', 'Gb']:
                self.addSZ('Mb')
        else:
            self.v = str(self.getValueBest(v))
        self.b = self.getValue()*_MultiDict[self.getSZ()]

    def __repr__(self):
        return self.v

    def addSZ(self, sz):
        self.v += sz

    def setSZ(self, sz):
        self.v[-2] = sz[0]
        self.v[-1] = sz[1]

    def getSZ(self):
        return self.v[-2:]

    def getValue(self):
        try:
            return float(self.v[:-2])
        except:
            return 0.0

    def getValueBytes(self):
        return self.b

    def getValueKb(self):
        sz = self.getSZ()
        if sz == 'Gb':
            return round(self.getValue()*1024.0*1024.0, 2)
        elif sz == 'Mb':
            return round(self.getValue()*1024.0, 2)
        return self.getValue()

    def getValueMb(self):
        sz = self.getSZ().lower()
        if sz == 'Kb':
            return round(self.getValue()/1024.0, 2)
        elif sz == 'Gb':
            return round(self.getValue()*1024.0, 2)
        return self.getValue()

    def getValueGb(self):
        sz = self.getSZ()
        if sz == 'Kb':
            return round(self.getValue()/(1024.0*1024.0), 2)
        elif sz == 'Mb':
            return round(self.getValue()/1024.0, 2)
        return self.getValue()

    def getValueBest(self, value=None):
        if value is not None:
            v = value
        else:
            v = self.getValueBytes()
        if v > _MultiDict['Gb']:
            res = round(v/_MultiDict['Gb'], 2)
            sz = 'Gb'
        elif v > _MultiDict['Mb']:
            res = round(v/_MultiDict['Mb'], 2)
            sz = 'Mb'
        else:
            res = round(v/_MultiDict['Kb'], 2)
            sz = 'Kb'
        return str(res) + sz


def SuffixIsCorrect(suffix):
    """
    Check input string to be a valid unit label.

    See global variable ``_Suffixes``.
    """
    global _Suffixes
    return suffix in list(_Suffixes.keys())


def SuffixLabels():
    """
    Return the correct suffix labels.
    """
    global _SuffixLabels
    return _SuffixLabels


def SameSuffix(suf1, suf2):
    """
    Compare 2 unit labels.

    Return True if both are same unit:

    from diskspace import SameSuffix

    SameSuffix('b','bytes')
    True
    """
    global _Suffixes
    if not SuffixIsCorrect(suf1):
        return False
    if not SuffixIsCorrect(suf2):
        return False
    return _Suffixes[suf1] == _Suffixes[suf2]


def MakeString(value, suf):
    """
    Method to join value and unit measure.
    """
    if not SuffixIsCorrect(suf):
        return str(value)
    if round(float(value)) == float(value):
        return str(int(float(value))) + ' ' + suf
    return str(value) + ' ' + suf


def SplitString(s):
    """
    Return tuple (<number>, <suffix>) or (None, None).

    from diskspace import SplitString

    SplitString("342.67Mb")
    (342.67, 'Mb')
    """
    num = s.rstrip('bytesBYTESgmkGMK ')
    suf = s.lstrip('0123456789., ').strip()
    try:
        num = float(num)
    except:
        return (None, None)

    if round(num) == num:
        num = int(num)

    if not SuffixIsCorrect(suf):
        return (None, None)

    return (num, suf)


def MakeStringFromBytes(value):
    """
    Make a correct string value with best units measure from given number of
    bytes.

    from diskspace import MakeStringFromBytes

    MakeStringFromBytes(123456)
    '120.56KB'

    MakeStringFromBytes(123.456789)
    '123 bytes'
    """
    try:
        v = float(value)
    except:
        return value
    if v > _Suffixes['GB']:
        res = round(v/_Suffixes['GB'], 2)
        sz = 'GB'
    elif v > _Suffixes['MB']:
        res = round(v/_Suffixes['MB'], 2)
        sz = 'MB'
    elif v > _Suffixes['KB']:
        res = round(v/_Suffixes['KB'], 2)
        sz = 'KB'
    else:
        res = int(v)
        sz = 'bytes'
    return MakeString(res, sz)


def GetBytesFromString(s, default=None):
    """
    Convert a string to a value in bytes, this is reverse method for
    MakeStringFromBytes.

    from diskspace import GetBytesFromString

    GetBytesFromString("123.456 Mb")
    129452998
    """
    num, suf = SplitString(s)
    if num is None:
        return default
    return int(num*_Suffixes[suf])


def MakeStringWithSuffix(s, suffix):
    """
    You can move strings from one unit measure to another.

    Convert input string to a string with given suffix.

    from diskspace import MakeStringWithSuffix

    MakeStringWithSuffix("12.345 Mb", "Kb")
    '12641.28 Kb'
    """
    b = GetBytesFromString(s)
    if b is None:
        return s
    if not SuffixIsCorrect(suffix):
        return s
    res = round(b/_Suffixes[suffix], 2)
    if _Suffixes[suffix] == 1.0:
        res = int(res)
    return MakeString(res, suffix)


def GetMegaBytesFromString(s):
    """
    This is just a wrapper for ``GetBytesFromString``, but return value in
    Megabytes.
    """
    b = GetBytesFromString(s)
    if b is None:
        return None
    return round(b/(1024*1024), 2)


def MakeStringFromString(s):
    """
    This method can be used during loading or checking user input.

    Call ``SplitString`` and than ``MakeString`` to "recreate" input
    string.
    """
    value, suf = SplitString(s)
    if value is None:
        return s
    return MakeString(value, suf)



#!/usr/bin/python
# service_gateway.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_gateway.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_gateway
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return GatewayService()


class GatewayService(LocalService):

    service_name = 'service_gateway'
    config_path = 'services/gateway/enabled'
    start_suspended = True

    def dependent_on(self):
        return [
            'service_network',
        ]

    def installed(self):
        from bitdust.userid import my_id
        if not my_id.isLocalIdentityReady():
            return False
        return True

    def network_configuration(self):
        from bitdust.crypt import key
        return {
            'session_key_type': key.SessionKeyType(),
        }

    def start(self):
        from bitdust.transport import packet_out
        from bitdust.transport import packet_in
        from bitdust.transport import gateway
        packet_out.init()
        packet_in.init()
        gateway.init()
        return True

    def stop(self):
        from bitdust.transport import packet_out
        from bitdust.transport import packet_in
        from bitdust.transport import gateway
        gateway.stop()
        gateway.shutdown()
        packet_out.shutdown()
        packet_in.shutdown()
        return True

    def on_suspend(self, *args, **kwargs):
        from bitdust.transport import gateway
        return gateway.stop()

    def on_resume(self, *args, **kwargs):
        from bitdust.transport import gateway
        if kwargs.get('cold_start') is True:
            return gateway.cold_start()
        return gateway.start()



#!/usr/bin/python
# config_defaults.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (config_types.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: config_types
"""

TYPE_UNDEFINED = 0
TYPE_BOOLEAN = 1
TYPE_STRING = 2
TYPE_TEXT = 3
TYPE_INTEGER = 4
TYPE_POSITIVE_INTEGER = 5
TYPE_NON_ZERO_POSITIVE_INTEGER = 6
TYPE_FOLDER_PATH = 7
TYPE_FILE_PATH = 8
TYPE_COMBO_BOX = 9
TYPE_PASSWORD = 10
TYPE_DISK_SPACE = 11
TYPE_PORT_NUMBER = 12
TYPE_FLOATING_POINT = 13
TYPE_POSITIVE_FLOATING_POINT = 14
TYPE_NON_ZERO_POSITIVE_FLOATING_POINT = 15


def labels():
    return {
        TYPE_UNDEFINED: 'undefined',
        TYPE_BOOLEAN: 'boolean',
        TYPE_STRING: 'string',
        TYPE_TEXT: 'text',
        TYPE_INTEGER: 'integer',
        TYPE_POSITIVE_INTEGER: 'positive integer',
        TYPE_NON_ZERO_POSITIVE_INTEGER: 'non zero positive integer',
        TYPE_FOLDER_PATH: 'folder path',
        TYPE_FILE_PATH: 'file path',
        TYPE_COMBO_BOX: 'selection',
        TYPE_PASSWORD: 'password',
        TYPE_DISK_SPACE: 'disk space',
        TYPE_PORT_NUMBER: 'port number',
        TYPE_FLOATING_POINT: 'floating point',
        TYPE_POSITIVE_FLOATING_POINT: 'non negative floating point',
        TYPE_NON_ZERO_POSITIVE_FLOATING_POINT: 'non zero positive floating point',
    }


def defaults():
    return {
        'interface/api/auth-secret-enabled': TYPE_BOOLEAN,
        'interface/api/rest-http-enabled': TYPE_BOOLEAN,
        'interface/api/rest-http-port': TYPE_PORT_NUMBER,
        'interface/api/web-socket-enabled': TYPE_BOOLEAN,
        'interface/api/web-socket-port': TYPE_PORT_NUMBER,
        'interface/ftp/enabled': TYPE_BOOLEAN,
        'interface/ftp/port': TYPE_PORT_NUMBER,
        'logs/api-enabled': TYPE_BOOLEAN,
        'logs/automat-events-enabled': TYPE_BOOLEAN,
        'logs/automat-transitions-enabled': TYPE_BOOLEAN,
        'logs/debug-level': TYPE_POSITIVE_INTEGER,
        'logs/memdebug-enabled': TYPE_BOOLEAN,
        'logs/memdebug-port': TYPE_PORT_NUMBER,
        'logs/memprofile-enabled': TYPE_BOOLEAN,
        'logs/packet-enabled': TYPE_BOOLEAN,
        'logs/stream-enabled': TYPE_BOOLEAN,
        'logs/stream-port': TYPE_PORT_NUMBER,
        'logs/traffic-enabled': TYPE_BOOLEAN,
        'logs/traffic-port': TYPE_PORT_NUMBER,

        # 'other/upnp-at-startup': TYPE_BOOLEAN,
        'paths/backups': TYPE_FOLDER_PATH,
        'paths/customers': TYPE_FOLDER_PATH,
        'paths/messages': TYPE_FOLDER_PATH,
        'paths/receipts': TYPE_FOLDER_PATH,
        'paths/restore': TYPE_FOLDER_PATH,
        'personal/private-key-size': TYPE_POSITIVE_INTEGER,
        'services/accountant/enabled': TYPE_BOOLEAN,
        'services/backup-db/enabled': TYPE_BOOLEAN,
        'services/backups/block-size': TYPE_DISK_SPACE,
        'services/backups/enabled': TYPE_BOOLEAN,
        'services/backups/keep-local-copies-enabled': TYPE_BOOLEAN,
        'services/backups/max-block-size': TYPE_DISK_SPACE,
        'services/backups/max-copies': TYPE_POSITIVE_INTEGER,
        'services/backups/wait-suppliers-enabled': TYPE_BOOLEAN,
        'services/blockchain-id/enabled': TYPE_BOOLEAN,
        'services/blockchain-authority/enabled': TYPE_BOOLEAN,
        'services/blockchain-authority/registration-bonus-coins': TYPE_POSITIVE_INTEGER,
        'services/blockchain-authority/requests-reading-offset': TYPE_POSITIVE_INTEGER,
        'services/blockchain-authority/requests-reading-limit': TYPE_NON_ZERO_POSITIVE_INTEGER,
        'services/blockchain-explorer/enabled': TYPE_BOOLEAN,
        'services/blockchain-explorer/host': TYPE_STRING,
        'services/blockchain-explorer/web-port': TYPE_PORT_NUMBER,
        'services/bismuth-blockchain/enabled': TYPE_BOOLEAN,
        'services/bismuth-node/enabled': TYPE_BOOLEAN,
        'services/bismuth-node/host': TYPE_STRING,
        'services/bismuth-node/tcp-port': TYPE_PORT_NUMBER,
        'services/bismuth-wallet/enabled': TYPE_BOOLEAN,
        'services/bismuth-pool/enabled': TYPE_BOOLEAN,
        'services/bismuth-pool/host': TYPE_STRING,
        'services/bismuth-pool/tcp-port': TYPE_PORT_NUMBER,
        'services/bismuth-miner/enabled': TYPE_BOOLEAN,
        'services/bismuth-miner/pool-host': TYPE_STRING,
        'services/bismuth-miner/pool-tcp-port': TYPE_PORT_NUMBER,
        'services/bismuth-identity/enabled': TYPE_BOOLEAN,
        'services/broadcasting/enabled': TYPE_BOOLEAN,
        'services/broadcasting/routing-enabled': TYPE_BOOLEAN,
        'services/broadcasting/max-broadcast-connections': TYPE_NON_ZERO_POSITIVE_INTEGER,
        'services/contract-chain/enabled': TYPE_BOOLEAN,
        'services/customer/enabled': TYPE_BOOLEAN,
        'services/customer/needed-space': TYPE_DISK_SPACE,
        'services/customer/suppliers-number': TYPE_COMBO_BOX,
        'services/customer-contracts/enabled': TYPE_BOOLEAN,
        'services/customer-family/enabled': TYPE_BOOLEAN,
        'services/customer-patrol/enabled': TYPE_BOOLEAN,
        'services/customer-patrol/customer-idle-days': TYPE_POSITIVE_INTEGER,
        'services/customer-support/enabled': TYPE_BOOLEAN,
        'services/data-disintegration/enabled': TYPE_BOOLEAN,
        'services/data-motion/enabled': TYPE_BOOLEAN,
        'services/data-motion/supplier-request-queue-size': TYPE_NON_ZERO_POSITIVE_INTEGER,
        'services/data-motion/supplier-sending-queue-size': TYPE_NON_ZERO_POSITIVE_INTEGER,
        'services/entangled-dht/enabled': TYPE_BOOLEAN,
        'services/entangled-dht/udp-port': TYPE_PORT_NUMBER,
        'services/entangled-dht/known-nodes': TYPE_STRING,
        'services/entangled-dht/attached-layers': TYPE_STRING,
        'services/employer/enabled': TYPE_BOOLEAN,
        'services/employer/replace-critically-offline-enabled': TYPE_BOOLEAN,
        'services/employer/candidates': TYPE_STRING,
        'services/gateway/enabled': TYPE_BOOLEAN,
        'services/gateway/p2p-timeout': TYPE_POSITIVE_INTEGER,
        'services/http-connections/enabled': TYPE_BOOLEAN,
        'services/http-connections/http-port': TYPE_PORT_NUMBER,
        'services/http-transport/enabled': TYPE_BOOLEAN,
        'services/http-transport/receiving-enabled': TYPE_BOOLEAN,
        'services/http-transport/sending-enabled': TYPE_BOOLEAN,
        'services/http-transport/priority': TYPE_POSITIVE_INTEGER,
        'services/identity-server/enabled': TYPE_BOOLEAN,
        'services/identity-server/host': TYPE_STRING,
        'services/identity-server/tcp-port': TYPE_PORT_NUMBER,
        'services/identity-server/web-port': TYPE_PORT_NUMBER,
        'services/identity-propagate/enabled': TYPE_BOOLEAN,
        'services/identity-propagate/known-servers': TYPE_STRING,
        'services/identity-propagate/preferred-servers': TYPE_STRING,
        'services/identity-propagate/min-servers': TYPE_NON_ZERO_POSITIVE_INTEGER,
        'services/identity-propagate/max-servers': TYPE_NON_ZERO_POSITIVE_INTEGER,
        'services/identity-propagate/automatic-rotate-enabled': TYPE_BOOLEAN,
        'services/identity-propagate/health-check-interval-seconds': TYPE_POSITIVE_INTEGER,
        'services/ip-port-responder/enabled': TYPE_BOOLEAN,
        'services/joint-postman/enabled': TYPE_BOOLEAN,
        'services/keys-registry/enabled': TYPE_BOOLEAN,
        'services/keys-storage/enabled': TYPE_BOOLEAN,
        'services/keys-storage/reset-unreliable-backup-copies': TYPE_BOOLEAN,
        'services/list-files/enabled': TYPE_BOOLEAN,
        'services/message-broker/enabled': TYPE_BOOLEAN,
        'services/message-broker/archive-chunk-size': TYPE_NON_ZERO_POSITIVE_INTEGER,
        'services/message-broker/message-ack-timeout': TYPE_NON_ZERO_POSITIVE_INTEGER,
        'services/message-broker/broker-negotiate-ack-timeout': TYPE_NON_ZERO_POSITIVE_INTEGER,
        'services/message-broker/preferred-brokers': TYPE_TEXT,
        'services/message-history/enabled': TYPE_BOOLEAN,
        'services/miner/enabled': TYPE_BOOLEAN,
        'services/my-data/enabled': TYPE_BOOLEAN,
        'services/my-ip-port/enabled': TYPE_BOOLEAN,
        'services/network/enabled': TYPE_BOOLEAN,
        'services/network/proxy/enabled': TYPE_BOOLEAN,
        'services/network/proxy/host': TYPE_STRING,
        'services/network/proxy/password': TYPE_PASSWORD,
        'services/network/proxy/port': TYPE_PORT_NUMBER,
        'services/network/proxy/ssl': TYPE_BOOLEAN,
        'services/network/proxy/username': TYPE_STRING,
        'services/network/receive-limit': TYPE_POSITIVE_INTEGER,
        'services/network/send-limit': TYPE_POSITIVE_INTEGER,
        'services/nodes-lookup/enabled': TYPE_BOOLEAN,
        'services/p2p-hookups/enabled': TYPE_BOOLEAN,
        'services/p2p-notifications/enabled': TYPE_BOOLEAN,
        'services/personal-messages/enabled': TYPE_BOOLEAN,
        'services/private-groups/enabled': TYPE_BOOLEAN,
        'services/private-groups/message-ack-timeout': TYPE_NON_ZERO_POSITIVE_INTEGER,
        'services/private-groups/broker-connect-timeout': TYPE_NON_ZERO_POSITIVE_INTEGER,
        'services/private-groups/preferred-brokers': TYPE_TEXT,
        'services/private-messages/enabled': TYPE_BOOLEAN,
        'services/private-messages/acknowledge-unread-messages-enabled': TYPE_BOOLEAN,
        'services/proxy-server/enabled': TYPE_BOOLEAN,
        'services/proxy-server/routes-limit': TYPE_POSITIVE_INTEGER,
        'services/proxy-server/current-routes': TYPE_TEXT,
        'services/proxy-transport/enabled': TYPE_BOOLEAN,
        'services/proxy-transport/sending-enabled': TYPE_BOOLEAN,
        'services/proxy-transport/receiving-enabled': TYPE_BOOLEAN,
        'services/proxy-transport/priority': TYPE_POSITIVE_INTEGER,
        'services/proxy-transport/my-original-identity': TYPE_TEXT,
        'services/proxy-transport/current-router': TYPE_STRING,
        'services/proxy-transport/preferred-routers': TYPE_TEXT,  # 'services/proxy-transport/router-lifetime-seconds': TYPE_POSITIVE_INTEGER,
        'services/rebuilding/enabled': TYPE_BOOLEAN,
        'services/restores/enabled': TYPE_BOOLEAN,
        'services/shared-data/enabled': TYPE_BOOLEAN,
        'services/supplier/donated-space': TYPE_DISK_SPACE,
        'services/supplier/enabled': TYPE_BOOLEAN,
        'services/supplier-contracts/enabled': TYPE_BOOLEAN,
        'services/supplier-contracts/initial-duration-hours': TYPE_NON_ZERO_POSITIVE_INTEGER,
        'services/supplier-contracts/duration-raise-factor': TYPE_NON_ZERO_POSITIVE_FLOATING_POINT,
        'services/tcp-connections/enabled': TYPE_BOOLEAN,
        'services/tcp-connections/tcp-port': TYPE_PORT_NUMBER,
        'services/tcp-connections/upnp-enabled': TYPE_BOOLEAN,
        'services/tcp-transport/enabled': TYPE_BOOLEAN,
        'services/tcp-transport/receiving-enabled': TYPE_BOOLEAN,
        'services/tcp-transport/sending-enabled': TYPE_BOOLEAN,
        'services/tcp-transport/priority': TYPE_POSITIVE_INTEGER,
        'services/udp-datagrams/enabled': TYPE_BOOLEAN,
        'services/udp-datagrams/udp-port': TYPE_PORT_NUMBER,
        'services/udp-transport/enabled': TYPE_BOOLEAN,
        'services/udp-transport/receiving-enabled': TYPE_BOOLEAN,
        'services/udp-transport/sending-enabled': TYPE_BOOLEAN,
        'services/udp-transport/priority': TYPE_POSITIVE_INTEGER,
    }
